import unittest


from pynusmv_tools.mucalculus.graph import Graph, domaintuple


class TestGraph(unittest.TestCase):
    
    def state_graph1(self):
        s1 = domaintuple(state="s1")
        s2 = domaintuple(state="s2")
        stay = domaintuple(act="stay")
        run = domaintuple(act="run")
        return Graph({s1: frozenset({(stay, s1), (run, s2)}),
                      s2: frozenset({(stay, s2)})})
    
    def state_graph2(self):
        s1 = domaintuple(state="s1")
        s3 = domaintuple(state="s3")
        stay = domaintuple(act="stay")
        run = domaintuple(act="run")
        return Graph({s1: frozenset({(stay, s1), (run, s3)}),
                      s3: frozenset({(stay, s3), (run, s1)})})
    
    def formula_graph(self):
        pq = domaintuple(formula="p & q")
        p = domaintuple(formula="p")
        q = domaintuple(formula="q")
        empty = domaintuple()
        return Graph({pq: frozenset({(empty, p), (empty, q)}),
                      p: frozenset(),
                      q: frozenset()})
    
    def test_union(self):
        states1 = self.state_graph1()
        states2 = self.state_graph2()
        formulas = self.formula_graph()
        
        self.assertEqual(states1, states1.union(states1))
        self.assertEqual(formulas, formulas.union(formulas))
        s1f = {}
        for state in states1.nodes:
            s1f[state] = states1[state]
        for formula in formulas.nodes:
            s1f[formula] = formulas[formula]
        self.assertEqual(states1.union(formulas),
                         Graph(s1f))
        
        s1 = domaintuple(state="s1")
        s2 = domaintuple(state="s2")
        s3 = domaintuple(state="s3")
        stay = domaintuple(act="stay")
        run = domaintuple(act="run")
        self.assertEqual(states1.union(states2),
                         Graph({s1: frozenset({(stay, s1),
                                               (run, s2),
                                               (run, s3)}),
                                s2: frozenset({(stay, s2)}),
                                s3: frozenset({(run, s1),
                                               (stay, s3)})}))
    
    def test_intersection(self):
        states1 = self.state_graph1()
        states2 = self.state_graph2()
        formulas = self.formula_graph()
        
        self.assertEqual(states1, states1.intersection(states1))
        self.assertEqual(formulas, formulas.intersection(formulas))
        self.assertEqual(states1.intersection(formulas),
                         Graph({}))
        
        self.assertEqual(states1.intersection(formulas.union(states1)),
                         states1)
        self.assertEqual(formulas.intersection(formulas.union(states1)),
                         formulas)
        
        s1 = domaintuple(state="s1")
        s2 = domaintuple(state="s2")
        s3 = domaintuple(state="s3")
        stay = domaintuple(act="stay")
        run = domaintuple(act="run")
        self.assertEqual(states1.intersection(states2),
                         Graph({s1: frozenset({(stay, s1)})}))
    
    def test_node_difference(self):
        states1 = self.state_graph1()
        states2 = self.state_graph2()
        formulas = self.formula_graph()
        
        self.assertEqual(Graph({}), states1.node_difference(states1))
        self.assertEqual(Graph({}),
                         formulas.node_difference(formulas))
        self.assertEqual(states1.node_difference(formulas),
                         states1)
        self.assertEqual(formulas.node_difference(states1), formulas)
        
        s1 = domaintuple(state="s1")
        s2 = domaintuple(state="s2")
        s3 = domaintuple(state="s3")
        stay = domaintuple(act="stay")
        run = domaintuple(act="run")
        self.assertEqual(states1.node_difference(states2),
                         Graph({s2: frozenset({(stay, s2)})}))
        self.assertEqual(states2.node_difference(states1),
                         Graph({s3: frozenset({(stay, s3)})}))
    
    def test_edge_difference(self):
        states1 = self.state_graph1()
        states2 = self.state_graph2()
        formulas = self.formula_graph()
        
        self.assertEqual(Graph({node: frozenset() for node in states1}),
                         states1.edge_difference(states1))
        self.assertEqual(Graph({node: frozenset() for node in formulas}),
                         formulas.edge_difference(formulas))
        self.assertEqual(states1.edge_difference(formulas),
                         states1)
        self.assertEqual(formulas.edge_difference(states1), formulas)
        
        s1 = domaintuple(state="s1")
        s2 = domaintuple(state="s2")
        s3 = domaintuple(state="s3")
        stay = domaintuple(act="stay")
        run = domaintuple(act="run")
        self.assertEqual(states1.edge_difference(states2),
                         Graph({s1: frozenset({(run, s2),}),
                                s2: frozenset({(stay, s2)})}))
        self.assertEqual(states2.edge_difference(states1),
                         Graph({s1: frozenset({(run, s3)}),
                                s3: frozenset({(run, s1), (stay, s3)})}))
    
    def test_selection(self):
        states1 = self.state_graph1()
        s1 = domaintuple(state="s1")
        s2 = domaintuple(state="s2")
        s3 = domaintuple(state="s3")
        stay = domaintuple(act="stay")
        run = domaintuple(act="run")
        
        self.assertEqual(states1.selection(node_selector=
                                           lambda n: n.state != "s1"),
                         Graph({s2: frozenset({(stay, s2)})}))
        self.assertEqual(states1.selection(edge_selector=
                                           lambda e: e[1].act == "run"),
                         Graph({s1: frozenset({(run, s2)}),
                                s2: frozenset()}))
        self.assertEqual(states1.selection(node_selector=
                                           lambda n: n.state == "s1",
                                           edge_selector=
                                           lambda e: e[1].act == "run"),
                         Graph({s1: frozenset()}))
    
    def test_join(self):
        states1 = self.state_graph1()
        states2 = self.state_graph2()
        formulas = self.formula_graph()
        
        self.assertEqual(states1.join(states1), states1)
        self.assertEqual(states1.join(states2), states1.intersection(states2))
        
        s1pq = domaintuple(state="s1", formula="p & q")
        s1p = domaintuple(state="s1", formula="p")
        s1q = domaintuple(state="s1", formula="q")
        s2pq = domaintuple(state="s2", formula="p & q")
        s2p = domaintuple(state="s2", formula="p")
        s2q = domaintuple(state="s2", formula="q")
        stay = domaintuple(act="stay")
        run = domaintuple(act="run")
        empty = domaintuple()
        j = Graph({s1pq: frozenset({(stay, s1pq), (stay, s1p), (stay, s1q),
                                    (run, s2pq), (run, s2p), (run, s2q),
                                    (empty, s1p), (empty, s1q)}),
                   s1p: frozenset({(stay, s1p), (run, s2p)}),
                   s1q: frozenset({(stay, s1q), (run, s2q)}),
                   s2pq: frozenset({(stay, s2pq), (stay, s2p), (stay, s2q),
                                    (empty, s2p), (empty, s2q)}),
                   s2p: frozenset({(stay, s2p)}),
                   s2q: frozenset({(stay, s2q)})})
        self.assertEqual(states1.join(formulas), j)
        self.assertEqual(j.join(states1), j)
        self.assertEqual(j.join(formulas), j)
    
    def test_projection(self):
        states1 = self.state_graph1()
        states2 = self.state_graph2()
        formulas = self.formula_graph()
        
        self.assertEqual(states1.projection(node_domain={"state"}), states1)
        self.assertEqual(states1.projection(edge_domain={"act"}), states1)
        self.assertEqual(states1.projection(node_domain={"state"},
                                            edge_domain={"act"}), states1)
        
        s1f = states1.join(formulas)
        s1 = domaintuple(state="s1")
        s2 = domaintuple(state="s2")
        s3 = domaintuple(state="s3")
        stay = domaintuple(act="stay")
        run = domaintuple(act="run")
        self.assertEqual(s1f.projection(node_domain={"state"},
                                        edge_domain={"act"}), states1)
        
        pq = domaintuple(formula="p & q")
        p = domaintuple(formula="p")
        q = domaintuple(formula="q")
        empty = domaintuple()
        self.assertEqual(s1f.projection(node_domain={"formula"},
                                        edge_domain={}),
                         Graph({pq: frozenset(),
                                p: frozenset(),
                                q: frozenset()}))
        self.assertEqual(s1f.projection(node_domain={"formula"}),
                         Graph({pq: frozenset({(stay, pq),
                                               (run, pq),
                                               (run, p),
                                               (run, q),
                                               (stay, p),
                                               (stay, q),
                                               (empty, p),
                                               (empty, q)}),
                                p: frozenset({(stay, p),
                                              (run, p)}),
                                q: frozenset({(stay, q),
                                              (run, q)})}))
    
    def test_extension(self):
        states1 = self.state_graph1()
        s1 = domaintuple(state="s1")
        s11 = domaintuple(state="s1", id=1)
        s2 = domaintuple(state="s2")
        s22 = domaintuple(state="s2", id=2)
        s3 = domaintuple(state="s3")
        stay = domaintuple(act="stay")
        staysure = domaintuple(act="stay", sure=True)
        run = domaintuple(act="run")
        runsure = domaintuple(act="run", sure=True)
        
        self.assertEqual(states1.extension(node_extension=lambda n:
                                           {"id": int(n.state[-1])}),
                         Graph({s11: frozenset({(stay, s11), (run, s22)}),
                                s22: frozenset({(stay, s22)})}))
        self.assertEqual(states1.extension(edge_extension=lambda e:
                                           {"sure": True}),
                         Graph({s1: frozenset({(staysure, s1),
                                               (runsure, s2)}),
                                s2: frozenset({(staysure, s2)})}))
        self.assertEqual(states1.extension(node_extension=lambda n:
                                           {"id": int(n.state[-1])},
                                           edge_extension=lambda e:
                                           {"sure": True}),
                         Graph({s11: frozenset({(staysure, s11),
                                                (runsure, s22)}),
                                s22: frozenset({(staysure, s22)})}))
    
    def test_mapping(self):
        states1 = self.state_graph1()
        s1 = domaintuple(state="s1")
        s1i = domaintuple(id=1)
        s2 = domaintuple(state="s2")
        s2i = domaintuple(id=2)
        s3 = domaintuple(state="s3")
        stay = domaintuple(act="stay")
        run = domaintuple(act="run")
        sure = domaintuple(sure=True)
        
        self.assertEqual(states1.mapping(node_mapping=lambda n:
                                         {"id": int(n.state[-1])}),
                         Graph({s1i: frozenset({(stay, s1i), (run, s2i)}),
                                s2i: frozenset({(stay, s2i)})}))
        self.assertEqual(states1.mapping(edge_mapping=lambda e:
                                         {"sure": True}),
                         Graph({s1: frozenset({(sure, s1), (sure, s2)}),
                                s2: frozenset({(sure, s2)})}))
        self.assertEqual(states1.mapping(node_mapping=lambda n:
                                         {"id": int(n.state[-1])},
                                         edge_mapping=lambda e:
                                         {"sure": True}),
                         Graph({s1i: frozenset({(sure, s1i), (sure, s2i)}),
                                s2i: frozenset({(sure, s2i)})}))
    
    def test_grouping(self):
        states1 = self.state_graph1()
        states2 = self.state_graph2()
        formulas = self.formula_graph()
        
        s1fo = domaintuple(state="s1",
                           formulas=frozenset({domaintuple(formula="p & q"),
                                               domaintuple(formula="p"),
                                               domaintuple(formula="q")}))
        s2fo = domaintuple(state="s2",
                           formulas=frozenset({domaintuple(formula="p & q"),
                                               domaintuple(formula="p"),
                                               domaintuple(formula="q")}))
        stayempty = domaintuple(actions=frozenset({domaintuple(act="stay"),
                                                   domaintuple()}))
        run = domaintuple(actions=frozenset({domaintuple(act="run")}))
        
        s1f = states1.join(formulas)
        g = s1f.grouping(node_group=("formulas", {"state"}),
                         edge_group=("actions", {}))
        self.assertEqual(g, Graph({s1fo: frozenset({(stayempty, s1fo),
                                                    (run, s2fo)}),
                                   s2fo: frozenset({(stayempty, s2fo)})}))
    
    def test_ungrouping(self):
        states1 = self.state_graph1()
        states2 = self.state_graph2()
        formulas = self.formula_graph()
        grouped = states1.join(formulas).grouping(node_group=("formulas",
                                                              {"state"}),
                                                  edge_group=("actions", {}))
        
        s1pq = domaintuple(state="s1", formula="p & q")
        s1p = domaintuple(state="s1", formula="p")
        s1q = domaintuple(state="s1", formula="q")
        s2pq = domaintuple(state="s2", formula="p & q")
        s2p = domaintuple(state="s2", formula="p")
        s2q = domaintuple(state="s2", formula="q")
        s1fo = domaintuple(state="s1",
                           formulas=frozenset({domaintuple(formula="p & q"),
                                               domaintuple(formula="p"),
                                               domaintuple(formula="q")}))
        s2fo = domaintuple(state="s2",
                           formulas=frozenset({domaintuple(formula="p & q"),
                                               domaintuple(formula="p"),
                                               domaintuple(formula="q")}))
        stay = domaintuple(act="stay")
        empty = domaintuple()
        run = domaintuple(act="run")
        stayempty = domaintuple(actions=frozenset({domaintuple(act="stay"),
                                                   domaintuple()}))
        arun = domaintuple(actions=frozenset({domaintuple(act="run")}))
        
        
        ungrouped = grouped.ungrouping(node_group="formulas")
        g = Graph({
                   s1pq: frozenset({(stayempty, s1pq),
                                    (stayempty, s1p),
                                    (stayempty, s1q),
                                    (arun, s2pq),
                                    (arun, s2p),
                                    (arun, s2q)}),
                   s1p: frozenset({(stayempty, s1pq),
                                   (stayempty, s1p),
                                   (stayempty, s1q),
                                   (arun, s2pq),
                                   (arun, s2p),
                                   (arun, s2q)}),
                   s1q: frozenset({(stayempty, s1pq),
                                   (stayempty, s1p),
                                   (stayempty, s1q),
                                   (arun, s2pq),
                                   (arun, s2p),
                                   (arun, s2q)}),
                   s2pq: frozenset({(stayempty, s2pq),
                                    (stayempty, s2p),
                                    (stayempty, s2q)}),
                   s2p: frozenset({(stayempty, s2pq),
                                   (stayempty, s2p),
                                   (stayempty, s2q)}),
                   s2q: frozenset({(stayempty, s2pq),
                                   (stayempty, s2p),
                                   (stayempty, s2q)})
                   })
        self.assertEqual(ungrouped, g)
        
        ungrouped = grouped.ungrouping(edge_group="actions")
        g = Graph({
                   s1fo: frozenset({(stay, s1fo),
                                    (empty, s1fo),
                                    (run, s2fo)}),
                   s2fo: frozenset({(stay, s2fo),
                                    (empty, s2fo)})
                   })
        self.assertEqual(ungrouped, g)
        
        ungrouped = grouped.ungrouping(node_group="formulas",
                                       edge_group="actions")
        g = Graph({
                   s1pq: frozenset({(stay, s1pq),
                                    (stay, s1p),
                                    (stay, s1q),
                                    (empty, s1pq),
                                    (empty, s1p),
                                    (empty, s1q),
                                    (run, s2pq),
                                    (run, s2p),
                                    (run, s2q)}),
                   s1p: frozenset({(stay, s1pq),
                                   (stay, s1p),
                                   (stay, s1q),
                                   (empty, s1pq),
                                   (empty, s1p),
                                   (empty, s1q),
                                   (run, s2pq),
                                   (run, s2p),
                                   (run, s2q)}),
                   s1q: frozenset({(stay, s1pq),
                                   (stay, s1p),
                                   (stay, s1q),
                                   (empty, s1pq),
                                   (empty, s1p),
                                   (empty, s1q),
                                   (run, s2pq),
                                   (run, s2p),
                                   (run, s2q)}),
                   s2pq: frozenset({(stay, s2pq),
                                    (stay, s2p),
                                    (stay, s2q),
                                    (empty, s2pq),
                                    (empty, s2p),
                                    (empty, s2q)}),
                   s2p: frozenset({(stay, s2pq),
                                   (stay, s2p),
                                   (stay, s2q),
                                   (empty, s2pq),
                                   (empty, s2p),
                                   (empty, s2q)}),
                   s2q: frozenset({(stay, s2pq),
                                   (stay, s2p),
                                   (stay, s2q),
                                   (empty, s2pq),
                                   (empty, s2p),
                                   (empty, s2q)})
                   })
        self.assertEqual(ungrouped, g)
    
    def test_ungrouping_not_domaintuple(self):
        states1 = self.state_graph1()
        states2 = self.state_graph2()
        formulas = self.formula_graph()
        grouped = states1.join(formulas).grouping(node_group=("formulas",
                                                              {"state"}),
                                                  edge_group=("actions", {}))
        def list_formulas(node):
            node = dict(node)
            node["formulas"] = frozenset({t.formula for t in node["formulas"]})
            return node
        mapped = grouped.mapping(node_mapping=list_formulas)
        
        s1pq = domaintuple(state="s1", formulas="p & q")
        s1p = domaintuple(state="s1", formulas="p")
        s1q = domaintuple(state="s1", formulas="q")
        s2pq = domaintuple(state="s2", formulas="p & q")
        s2p = domaintuple(state="s2", formulas="p")
        s2q = domaintuple(state="s2", formulas="q")
        stayempty = domaintuple(actions=frozenset({domaintuple(act="stay"),
                                                   domaintuple()}))
        arun = domaintuple(actions=frozenset({domaintuple(act="run")}))
        
        ungrouped = mapped.ungrouping(node_group="formulas")
        g = Graph({
                   s1pq: frozenset({(stayempty, s1pq),
                                    (stayempty, s1p),
                                    (stayempty, s1q),
                                    (arun, s2pq),
                                    (arun, s2p),
                                    (arun, s2q)}),
                   s1p: frozenset({(stayempty, s1pq),
                                   (stayempty, s1p),
                                   (stayempty, s1q),
                                   (arun, s2pq),
                                   (arun, s2p),
                                   (arun, s2q)}),
                   s1q: frozenset({(stayempty, s1pq),
                                   (stayempty, s1p),
                                   (stayempty, s1q),
                                   (arun, s2pq),
                                   (arun, s2p),
                                   (arun, s2q)}),
                   s2pq: frozenset({(stayempty, s2pq),
                                    (stayempty, s2p),
                                    (stayempty, s2q)}),
                   s2p: frozenset({(stayempty, s2pq),
                                   (stayempty, s2p),
                                   (stayempty, s2q)}),
                   s2q: frozenset({(stayempty, s2pq),
                                   (stayempty, s2p),
                                   (stayempty, s2q)})
                   })
        self.assertEqual(ungrouped, g)
    
    def test_ungrouping_not_iterable(self):
        states1 = self.state_graph1()
        ungrouped = states1.ungrouping(node_group="other")
        self.assertEqual(ungrouped, states1)
        
        def state_to_id(node):
            node = dict(node)
            node["state"] = int(node["state"][1:])
            return node
        states1 = states1.mapping(node_mapping=state_to_id)
        self.assertEqual(states1, states1.ungrouping(node_group="state"))


"""
The partial approach, with and without pre-filtering.
"""

import gc

from pynusmv.dd import BDD
from pynusmv.mc import eval_simple_expression
from pynusmv.utils import fixpoint

from ..atlkFO.ast import (TrueExp, FalseExp, Init, Reachable,
                          Atom, Not, And, Or, Implies, Iff, 
                          AF, AG, AX, AU, AW, EF, EG, EX, EU, EW,
                          nK, nE, nD, nC, K, E, D, C,
                          CEF, CEG, CEX, CEU, CEW, CAF, CAG, CAX, CAU, CAW)
from ..atlkFO.eval import (fair_states, ex, eg, eu, nk, ne, nd, nc)

from .common import *
from .utils import *

GC_FREQUENCE = 100

def evalATLK(mas, formula, states=None, pre_filtering=False):
    """
    Return the BDD representing the subset of the given states of mas
    satisfying formula.
    
    mas -- a multi-agents system;
    formula -- an AST-based ATLK formula;
    states -- if not None, a subset of states of mas;
              if None, the initial states are used instead;
    pre_filtering -- whether or not applying pre-filtering.
    """
    
    if states is None:
        states = mas.init
    
    if type(formula) is TrueExp:
        return states
        
    elif type(formula) is FalseExp:
        return BDD.false(mas)
        
    elif type(formula) is Init:
        return mas.init & states
        
    elif type(formula) is Reachable:
        return mas.reachable_states & states
    
    elif type(formula) is Atom:
        return eval_simple_expression(mas, formula.value) & states
        
    elif type(formula) is Not:
        return states - evalATLK(mas,
                                 formula.child,
                                 states=states,
                                 pre_filtering=pre_filtering)
        
    elif type(formula) is And:
        return (evalATLK(mas,
                         formula.left,
                         states=states,
                         pre_filtering=pre_filtering) &
                evalATLK(mas,
                         formula.right,
                         states=states,
                         pre_filtering=pre_filtering))
        
    elif type(formula) is Or:
        return (evalATLK(mas,
                         formula.left,
                         states=states,
                         pre_filtering=pre_filtering) |
                evalATLK(mas,
                         formula.right,
                         states=states,
                         pre_filtering=pre_filtering))
        
    elif type(formula) is Implies:
        # a -> b = ~a | b
        p = formula.left
        q = formula.right
        new_formula = Or(Not(p), q)
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is Iff:
        # a <-> b = (a & b) | (~a & ~b)
        p = formula.left
        q = formula.right
        new_formula = Or(And(p, q), And(Not(p), Not(q)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is EX:
        return ex(mas,
                  evalATLK(mas,
                           formula.child,
                           states=mas.post(states),
                           pre_filtering=pre_filtering)) & states
        
    elif type(formula) is AX:
        # AX p = ~EX ~p
        new_formula = Not(EX(Not(spec.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is EG:
        return eg(mas,
                  evalATLK(mas,
                           formula.child,
                           states=reach(mas, states),
                           pre_filtering=pre_filtering)) & states
        
    elif type(formula) is AG:
        # AG p = ~EF ~p = ~E[ true U ~p ]
        new_formula = Not(EF(Not(formula.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is EU:
        return eu(mas,
                  evalATLK(mas,
                           formula.left,
                           states=reach(mas, states),
                           pre_filtering=pre_filtering),
                  evalATLK(mas,
                           formula.right,
                           states=reach(mas, states),
                           pre_filtering=pre_filtering))
        
    elif type(formula) is AU:
        # A[p U q] = ~E[~q W ~p & ~q] = ~(E[~q U ~p & ~q] | EG ~q)
        p = formula.left
        q = formula.right
        new_formula = Not(Or(EU(Not(q), And(Not(p), Not(q))), EG(Not(q))))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is EF:
        # EF p = E[ true U p ]
        new_formula = EU(TrueExp(), formula.child)
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is AF:
        # AF p = ~EG ~p
        new_formula = Not(EG(Not(formula.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is EW:
        # E[ p W q ] = E[ p U q ] | EG p
        p = formula.left
        q = formula.right
        new_formula = Or(EU(p, q), EG(p))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is AW:
        # A[p W q] = ~E[~q U ~p & ~q]
        p = formula.left
        q = formula.right
        new_formula = Not(EU(Not(q), And(Not(p), Not(q))))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is nK:
        agent = [formula.agent.value]
        phi = evalATLK(mas,
                       formula.child,
                       states=mas.equivalent_states(states, agent) &
                              mas.reachable_states,
                       pre_filtering=pre_filtering)
        return nk(mas,
                  formula.agent.value,
                  phi) & states
        
    elif type(formula) is K:
        # K<'a'> p = ~nK<'a'> ~p
        new_formula = Not(nK(formula.agent, Not(formula.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is nE:
        agents = [a.value for a in formula.group]
        equiv = Eequiv(mas, agents, states)
        phi = evalATLK(mas,
                       formula.child,
                       states=equiv,
                       pre_filtering=pre_filtering)
        return ne(mas, agents, phi) & states
        
    elif type(formula) is E:
        # E<g> p = ~nE<g> ~p
        new_formula = Not(nE(formula.group, Not(formula.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is nD:
        agents = [a.value for a in formula.group]
        equiv = Dequiv(mas, agents, states)
        phi = evalATLK(mas,
                       formula.child,
                       states=equiv,
                       pre_filtering=pre_filtering)
        return nd(mas, agents, phi) & states
        
    elif type(formula) is D:
        # D<g> p = ~nD<g> ~p
        new_formula = Not(nD(formula.group, Not(formula.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is nC:
        agents = [a.value for a in formula.group]
        equiv = Cequiv(mas, agents, states)
        phi = evalATLK(mas,
                       formula.child,
                       states=equiv,
                       pre_filtering=pre_filtering)
        return nd(mas, agents, phi) & states
        
    elif type(formula) is C:
        # C<g> p = ~nC<g> ~p
        new_formula = Not(nC(formula.group, Not(formula.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
    
    elif type(formula) is CAX:
        # [g] X p = ~<g> X ~p
        new_formula = CEX(formula.group, Not(formula.child))
        return ~evalATLK(mas,
                         new_formula,
                         states=states,
                         pre_filtering=pre_filtering)
        
    elif type(formula) is CAG:
        # [g] G p = ~<g> F ~p
        new_formula = CEF(formula.group, Not(formula.child))
        return ~evalATLK(mas,
                         new_formula,
                         states=states,
                         pre_filtering=pre_filtering)
        
    elif type(formula) is CAU:
        # [g][p U q] = ~<g>[ ~q W ~p & ~q ]
        new_formula = CEW(formula.group,
                      Not(formula.right),
                      And(Not(formula.left), Not(formula.right)))
        return ~evalATLK(mas,
                         new_formula,
                         states=states,
                         pre_filtering=pre_filtering)
        
    elif type(formula) is CAF:
        # [g] F p = ~<g> G ~p
        new_formula = CEG(formula.group, Not(formula.child))
        return ~evalATLK(mas,
                         new_formula,
                         states=states,
                         pre_filtering=pre_filtering)
        
    elif type(formula) is CAW:
        # [g][p W q] = ~<g>[~q U ~p & ~q]
        new_formula = CEU(formula.group,
                      Not(formula.right),
                      And(Not(formula.left), Not(formula.right)))
        return ~evalATLK(mas,
                         new_formula,
                         states=states,
                         pre_filtering=pre_filtering)
        
    elif type(formula) is CEG:
        # <g> G p = <g> p W false
        new_formula = CEW(formula.group, formula.child, FalseExp())
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is CEF:
        # <g> F p = <g> true U p
        new_formula = CEU(formula.group, TrueExp(), formula.child)
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
                   
    elif type(formula) in {CEX, CEU, CEW}:
        return eval_strat(mas,
                          formula,
                          states,
                          pre_filtering=pre_filtering)
        
    else:
        raise Exception("evalATLK: unrecognized formula type:" + str(formula))


def eval_strat(mas, formula, states, pre_filtering=False):
    """
    Return the BDD representing the subset of the given states of mas
    satisfying formula.
    
    mas -- a multi-agents system;
    formula -- an AST-based ATLK strategic formula;
    states -- a subset of states of mas;
    pre_filtering -- whether or not applying pre-filtering.
    """
    assert(type(formula) in {CEX, CEU, CEW})
    
    agents = [atom.value for atom in formula.group]
    agents = agents_in_list(mas, agents)
    
    # pre-filter if needed
    if pre_filtering:
        if type(formula) is CEX:
            sub_states = post_through(mas,
                                      agents,
                                      Eequiv(mas, agents, states),
                                      mas.protocol(agents))
            sub = evalATLK(mas,
                           formula.child,
                           states=sub_states,
                           pre_filtering=pre_filtering)
            filtered = filter_cex_moves(mas, agents, sub, mas.protocol(agents))
        elif type(formula) is CEU:
            sub_1 = evalATLK(mas,
                             formula.left,
                             states=reach(mas, states),
                             pre_filtering=pre_filtering)
            sub_2 = evalATLK(mas,
                             formula.right,
                             states=reach(mas, states),
                             pre_filtering=pre_filtering)
            filtered = filter_ceu_moves(mas,
                                        agents,
                                        sub_1,
                                        sub_2,
                                        mas.protocol(agents))
        elif type(formula) is CEW:
            sub_1 = evalATLK(mas,
                             formula.left,
                             states=reach(mas, states),
                             pre_filtering=pre_filtering)
            sub_2 = evalATLK(mas,
                             formula.right,
                             states=reach(mas, states),
                             pre_filtering=pre_filtering)
            filtered = filter_cew_moves(mas,
                                        agents,
                                        sub_1,
                                        sub_2,
                                        mas.protocol(agents))
        else:
            raise Exception("eval_strat: unrecognized formula type:" +
                            str(formula))
        
        if (states & filtered).is_false():
            return BDD.false(mas)
    # otherwise consider the whole protocol
    else:
        filtered = mas.protocol(agents)
    
    # Remove states for which we already know there are no possible strategies
    states = (states & filtered).forsome(mas.bddEnc.inputsCube)
    
    # split and accumulate
    nb_strats = 0
    
    sat = BDD.false(mas)
    for strat in partial_strategies_filtered(mas,
                                             agents,
                                             Eequiv(mas, agents, states),
                                             filtered):
        if sat == states:
            return sat
        
        if type(formula) is CEX:
            sub_states = post_through(mas,
                                      agents,
                                      Eequiv(mas, agents, states),
                                      strat)
            sub = evalATLK(mas,
                           formula.child,
                           states=sub_states,
                           pre_filtering=pre_filtering)
            winning = filter_cex(mas, agents, sub, strat)
        elif type(formula) is CEU:
            sub_1 = evalATLK(mas,
                             formula.left,
                             states=strat.forsome(mas.bddEnc.inputsCube),
                             pre_filtering=pre_filtering)
            sub_2 = evalATLK(mas,
                             formula.right,
                             states=strat.forsome(mas.bddEnc.inputsCube),
                             pre_filtering=pre_filtering)
            winning = filter_ceu(mas, agents, sub_1, sub_2, strat)
        elif type(formula) is CEW:
            sub_1 = evalATLK(mas,
                             formula.left,
                             states=strat.forsome(mas.bddEnc.inputsCube),
                             pre_filtering=pre_filtering)
            sub_2 = evalATLK(mas,
                             formula.right,
                             states=strat.forsome(mas.bddEnc.inputsCube),
                             pre_filtering=pre_filtering)
            winning = filter_cew(mas, agents, sub_1, sub_2, strat)
        else:
            raise Exception("eval_strat: unrecognized formula type:" +
                            str(formula))
        
        sat |= all_equiv_sat(mas, agents, winning) & states
        
        nb_strats += 1
        if nb_strats % GC_FREQUENCE == 0:
            gc.collect()
    
    return sat


def reach_split_filtered(mas, agents, moves, filtered):
    """
    Return the set of largest non-agents-conflicting extensions of the given
    set of moves with moves of filtered reachable from moves.
    
    mas -- a multi-agents system;
    agents -- a subset of agents of mas;
    moves -- a subset of moves for agents;
    filtered -- a subset of moves for agents.
    """
    new_states = (post_through(mas, agents, BDD.true(mas), moves) -
                  moves.forsome(mas.bddEnc.inputsCube))
    new_moves = new_states & filtered
    compatible = compatible_moves(mas, agents, new_moves, moves)
    if compatible.is_false():
        yield moves
    else:
        for sub_strat in split(mas, agents, compatible):
            for new_strat in reach_split_filtered(mas,
                                                  agents,
                                                  moves | sub_strat,
                                                  filtered):
                yield new_strat


def partial_strategies_filtered(mas, agents, states, filtered):
    """
    Return the set of non-agents-conflicting subsets of filtered reachable from
    some of the given states in mas.
    
    mas -- a multi-agents system;
    agents -- a subset of agents of mas;
    states -- a subset of states of mas;
    filtered -- a subset of moves for agents.
    """
    for nc_moves in split(mas, agents, states & filtered):
        for strat in reach_split_filtered(mas, agents, nc_moves, filtered):
            yield strat


# ----- caching ---------------------------------------------------------------

__evalATLK_cache = {}
__orig_evalATLK = evalATLK
def __cached_evalATLK(mas, formula, states=None, pre_filtering=False):
    if states is None:
        states = mas.init
    
    if (mas, formula) in __evalATLK_cache.keys():
        sat, unsat = __evalATLK_cache[(mas, formula)]
    else:
        false = BDD.false(mas)
        sat, unsat = false, false
    
    remaining = states - (sat | unsat)
    
    if remaining.isnot_false():
        remsat = __orig_evalATLK(mas,
                                 formula,
                                 states=remaining,
                                 pre_filtering=pre_filtering)
        remunsat = remaining - remsat
        __evalATLK_cache[(mas, formula)] = (sat + remsat, unsat + remunsat)
    else:
        remsat = BDD.false(mas)
    
    return sat + remsat

evalATLK = __cached_evalATLK



"""

The mucalculus package defines mu-calculus models, formulas, adequate
explanations and generates adequate explanations.

The package is structured as follows:

    - model.py defines mu-calculus models and provides a way to build such a
      model from the currently read model;
    - formula.py defines the different mu-calculus operators; each operator
      defines its own evaluation and explanation functions;
    - explanation.py defines the structure of adequate explanations for the
      mu-calculus. These explanations are produced by explain functions of
      mu-calculus operators.
"""

from copy import deepcopy
from collections import Mapping, MutableSet
from pynusmv.exception import PyNuSMVError

class MuCalculusError(PyNuSMVError):
    """Generic mu-calculus package error."""
    pass

class MuCalculusModelError(MuCalculusError):
    """Model-related errors."""
    pass

class UnknownTransitionError(MuCalculusError):
    """Unknown transition name."""
    pass

class UnknownVariableError(MuCalculusError):
    """Variable not in context."""
    pass

class CannotExplainFalseError(MuCalculusError):
    """Trying to explain why state belongs to false."""
    pass

class TranslationError(MuCalculusError):
    """Cannot translate model or formula."""
    pass

class ChoiceError(MuCalculusError):
    """A choice is incorrect."""
    pass

class ExclusiveChoiceError(ChoiceError):
    """The choice must be exclusive but was not."""
    pass


class hashabledict(Mapping):
    """
    Hashable read-only dictionary.
    
    Inspired from frozendict https://pypi.python.org/pypi/frozendict/.
    """

    def __init__(self, *args, **kwargs):
        self.__dict = dict(*args, **kwargs)
        self.__hash = None

    def __getitem__(self, key):
        return self.__dict[key]

    def copy(self):
        return hashabledict(self)

    def __iter__(self):
        return iter(self.__dict)

    def __len__(self):
        return len(self.__dict)

    def __repr__(self):
        return '<hashabledict %s>' % repr(self.__dict)

    def __hash__(self):
        if self.__hash is None:
            self.__hash = hash(frozenset(self.__dict.items()))
        return self.__hash


class orderedset(MutableSet):
    """
    Ordered set. A classic set that remembers order of insertion.
    """

    def __init__(self, iterable=None):
        self.__set = set()
        self.__list = list()
        if iterable is not None:
            for e in iterable:
                self.add(e)

    def __contains__(self, element):
        return element in self.__set

    def __iter__(self):
        return iter(self.__list)

    def __len__(self):
        return len(self.__set)

    def add(self, element):
        if element not in self.__set:
            self.__set.add(element)
            self.__list.append(element)

    def discard(self, element):
        if element in self.__set:
            self.__set.discard(element)
            self.__list.remove(element)

    def __repr__(self):
        return '<orderedset {%s}>' % ", ".join(str(elem) for elem in self)


"""
The naive approach, with and without pre-filtering.
"""

import operator
from functools import reduce
from collections import OrderedDict

from pynusmv.dd import BDD
from pynusmv.fsm import BddTrans
from pynusmv.mc import eval_simple_expression
from pynusmv.utils import fixpoint
from pynusmv import model
from pynusmv import glob
from pynusmv import node

from ..atlkFO.ast import (TrueExp, FalseExp, Init, Reachable,
                          Atom, Not, And, Or, Implies, Iff, 
                          AF, AG, AX, AU, AW, EF, EG, EX, EU, EW,
                          nK, nE, nD, nC, K, E, D, C,
                          CEF, CEG, CEX, CEU, CEW, CAF, CAG, CAX, CAU, CAW)
from ..atlkFO.eval import (fair_states, nk, ne, nd, nc)

from .common import *
from .utils import *

def evalATLK(mas, formula, pre_filtering=False):
    """
    Return the BDD representing the set of states of mas satisfying formula.
    
    mas -- a multi-agents system;
    formula -- an AST-based ATLK formula;
    pre_filtering -- whether or not applying pre-filtering.
    """
    
    if type(formula) is TrueExp:
        return BDD.true(mas)
        
    elif type(formula) is FalseExp:
        return BDD.false(mas)
        
    elif type(formula) is Init:
        return mas.init
        
    elif type(formula) is Reachable:
        return mas.reachable_states
    
    elif type(formula) is Atom:
        return eval_simple_expression(mas, formula.value)
        
    elif type(formula) is Not:
        return ~evalATLK(mas, formula.child, pre_filtering=pre_filtering)
        
    elif type(formula) is And:
        return (evalATLK(mas, formula.left, pre_filtering=pre_filtering) &
                evalATLK(mas, formula.right, pre_filtering=pre_filtering))
        
    elif type(formula) is Or:
        return (evalATLK(mas, formula.left, pre_filtering=pre_filtering) |
                evalATLK(mas, formula.right, pre_filtering=pre_filtering))
        
    elif type(formula) is Implies:
        # a -> b = ~a | b
        return ((~evalATLK(mas, formula.left, pre_filtering=pre_filtering)) |
                evalATLK(mas, formula.right, pre_filtering=pre_filtering))
        
    elif type(formula) is Iff:
        # a <-> b = (a & b) | (~a & ~b)
        l = evalATLK(mas, formula.left, pre_filtering=pre_filtering)
        r = evalATLK(mas, formula.right, pre_filtering=pre_filtering)
        return (l & r) | ((~l) & (~r))
        
    elif type(formula) is EX:
        return ex(mas,
                  evalATLK(mas, formula.child, pre_filtering=pre_filtering))
        
    elif type(formula) is AX:
        # AX p = ~EX ~p
        return ~ex(mas,
                   ~evalATLK(mas, formula.child, pre_filtering=pre_filtering))
        
    elif type(formula) is EG:
        # EG p = E p W false
        new_formula = EW(formula.child, FalseExp())
        return evalATLK(mas, new_formula, pre_filtering=pre_filtering)
        
    elif type(formula) is AG:
        # AG p = ~EF ~p = ~E[ true U ~p ]
        new_formula = EF(Not(formula.child))
        return ~evalATLK(mas, new_formula, pre_filtering=pre_filtering)
        
    elif type(formula) is EU:
        return eu(mas,
                  evalATLK(mas, formula.left, pre_filtering=pre_filtering),
                  evalATLK(mas, formula.right, pre_filtering=pre_filtering))
        
    elif type(formula) is AU:
        # A[p U q] = ~E[~q W ~p & ~q] = ~(E[~q U ~p & ~q] | EG ~q)
        new_formula = EW(Not(formula.right),
                         And(Not(formula.left), Not(formula.right)))
        return ~evalATLK(mas, new_formula, pre_filtering=pre_filtering)
        
    elif type(formula) is EF:
        # EF p = E true U p
        new_formula = EU(TrueExp(), formula.child)
        return evalATLK(mas, new_formula, pre_filtering=pre_filtering)
        
    elif type(formula) is AF:
        # AF p = ~EG ~p
        new_formula = EG(Not(formula.child))
        return ~evalATLK(mas, new_formula, pre_filtering=pre_filtering)
        
    elif type(formula) is EW:
        return ew(mas,
                  evalATLK(mas, formula.left, pre_filtering=pre_filtering),
                  evalATLK(mas, formula.right, pre_filtering=pre_filtering))
        
    elif type(formula) is AW:
        # A[p W q] = ~E[~q U ~p & ~q]
        p = evalATLK(mas, formula.left, pre_filtering=pre_filtering)
        q = evalATLK(mas, formula.right, pre_filtering=pre_filtering)
        return ~eu(mas, ~q, ~p & ~q)
        
    elif type(formula) is nK:
        return nk(mas,
                  formula.agent.value,
                  evalATLK(mas, formula.child, pre_filtering=pre_filtering))
        
    elif type(formula) is K:
        # K<'a'> p = ~nK<'a'> ~p
        return ~nk(mas,
                   formula.agent.value,
                   ~evalATLK(mas, formula.child, pre_filtering=pre_filtering))
        
    elif type(formula) is nE:
        return ne(mas,
                  [a.value for a in formula.group],
                  evalATLK(mas, formula.child, pre_filtering=pre_filtering))
        
    elif type(formula) is E:
        # E<g> p = ~nE<g> ~p
        return ~ne(mas,
                   [a.value for a in formula.group],
                   ~evalATLK(mas, formula.child, pre_filtering=pre_filtering))
        
    elif type(formula) is nD:
        return nd(mas,
                  [a.value for a in formula.group],
                  evalATLK(mas, formula.child, pre_filtering=pre_filtering))
        
    elif type(formula) is D:
        # D<g> p = ~nD<g> ~p
        return ~nd(mas,
                   [a.value for a in formula.group],
                   ~evalATLK(mas, formula.child, pre_filtering=pre_filtering))
        
    elif type(formula) is nC:
        return nc(mas,
                  [a.value for a in formula.group],
                  evalATLK(mas, formula.child, pre_filtering=pre_filtering))
        
    elif type(formula) is C:
        # C<g> p = ~nC<g> ~p
        return ~nc(mas,
                   [a.value for a in formula.group],
                   ~evalATLK(mas, formula.child, pre_filtering=pre_filtering))
    
    elif type(formula) is CAX:
        # [g] X p = ~<g> X ~p
        new_formula = CEX(formula.group, Not(formula.child))
        return ~evalATLK(mas, new_formula, pre_filtering=pre_filtering)
        
    elif type(formula) is CAG:
        # [g] G p = ~<g> F ~p
        new_formula = CEF(formula.group, Not(formula.child))
        return ~evalATLK(mas, new_formula, pre_filtering=pre_filtering)
        
    elif type(formula) is CAU:
        # [g][p U q] = ~<g>[ ~q W ~p & ~q ]
        new_formula = CEW(formula.group,
                      Not(formula.right),
                      And(Not(formula.left), Not(formula.right)))
        return ~evalATLK(mas, new_formula, pre_filtering=pre_filtering)
        
    elif type(formula) is CAF:
        # [g] F p = ~<g> G ~p
        new_formula = CEG(formula.group, Not(formula.child))
        return ~evalATLK(mas, new_formula, pre_filtering=pre_filtering)
        
    elif type(formula) is CAW:
        # [g][p W q] = ~<g>[~q U ~p & ~q]
        new_formula = CEU(formula.group,
                      Not(formula.right),
                      And(Not(formula.left), Not(formula.right)))
        return ~evalATLK(mas, new_formula, pre_filtering=pre_filtering)
        
    elif type(formula) is CEG:
        # <g> G p = <g> p W false
        new_formula = CEW(formula.group, formula.child, FalseExp())
        return evalATLK(mas, new_formula, pre_filtering=pre_filtering)
        
    elif type(formula) is CEF:
        # <g> F p = <g> true U p
        new_formula = CEU(formula.group, TrueExp(), formula.child)
        return evalATLK(mas, new_formula, pre_filtering=pre_filtering)
                   
    elif type(formula) in {CEX, CEU, CEW}:
        return eval_strat(mas, formula, pre_filtering=pre_filtering)
        
    else:
        raise Exception("evalATLK: unrecognized formula type:" + str(formula))


def eval_strat(mas, formula, pre_filtering=False):
    """
    Return the BDD representing the set of states of mas satisfying formula.
    
    mas -- a multi-agents system;
    formula -- an AST-based ATLK strategic formula;
    pre_filtering -- whether or not applying pre-filtering.
    """
    assert(type(formula) in {CEX, CEU, CEW})
    
    agents = [atom.value for atom in formula.group]
    agents = agents_in_list(mas, agents)
    
    # pre-filter if needed
    if pre_filtering:
        if type(formula) is CEX:
            sub = evalATLK(mas, formula.child, pre_filtering=pre_filtering)
            filtered = filter_cex_moves(mas, agents, sub, mas.protocol(agents))
        elif type(formula) is CEU:
            sub_1 = evalATLK(mas, formula.left, pre_filtering=pre_filtering)
            sub_2 = evalATLK(mas, formula.right, pre_filtering=pre_filtering)
            filtered = filter_ceu_moves(mas,
                                        agents,
                                        sub_1,
                                        sub_2,
                                        mas.protocol(agents))
        elif type(formula) is CEW:
            sub_1 = evalATLK(mas, formula.left, pre_filtering=pre_filtering)
            sub_2 = evalATLK(mas, formula.right, pre_filtering=pre_filtering)
            filtered = filter_cew_moves(mas,
                                        agents,
                                        sub_1,
                                        sub_2,
                                        mas.protocol(agents))
        else:
            raise Exception("eval_strat: unrecognized formula type:" +
                            str(formula))
        
        if filtered.is_false():
            return filtered
    # otherwise consider the whole protocol
    else:
        filtered = mas.protocol(agents)
    
    # Encode strategies
    encode_strategies(mas, agents, formula, filtered)

    # Compute the set of winning states
    if type(formula) is CEX:
        sub = evalATLK(mas, formula.child, pre_filtering=pre_filtering)
        winning = eval_cex(mas, formula, agents, sub)
    elif type(formula) is CEU:
        sub_1 = evalATLK(mas, formula.left, pre_filtering=pre_filtering)
        sub_2 = evalATLK(mas, formula.right, pre_filtering=pre_filtering)
        winning = eval_ceu(mas, formula, agents, sub_1, sub_2)
    elif type(formula) is CEW:
        sub_1 = evalATLK(mas, formula.left, pre_filtering=pre_filtering)
        sub_2 = evalATLK(mas, formula.right, pre_filtering=pre_filtering)
        winning = eval_cew(mas, formula, agents, sub_1, sub_2)
    else:
        raise Exception("eval_strat: unrecognized formula type:" +
                        str(formula))
    
    winning = (winning.forsome(mas.bddEnc.inputsCube) & mas.reachable_states)
    return winning


# ----- Encoding strategies ---------------------------------------------------

def encode_strategies(mas, agents, formula, filtered):
    """
    Encode the strategies for agents, to check formula, given the filtered
    moves.
    
    mas -- a multi-agents system;
    agents -- a subset of agents of mas;
    formula -- an ATLK_irF strategic formula where the group is agents;
    filtered -- either the complete protocol of the agents, or the filtered
                moves for agents, for the given formula.
    
    The result is the presence, in mas.transitions[formula], of three relations
    under the "jump", "equiv", and "follow" keys, to perform the model
    checking:
        * the "jump" relation encode the fact that the original state stays the
          same and all strategies can change;
        * the "equiv" relation encode the fact that the states are equivalent
          for some agent in the group of formula, and the strategies of these
          agents are kept the same;
        * the "follow" relation encode the original transition relation with
          each agent of the group of formula following his strategy encoded
          in the current state (and that these strategies are kept the same).
    
    Note: mas.encoded is populated with intermediate cached information:
        * mas.encoded[(agent, filtered)] gives two (model-based) relations:
          + the relation encoding the fact that the encoded strategies of agent
            from filtered are kept the same;
          + the relation encoding the fact that the encoded strategies of agent
            are followed;
          they only depends on the agent and the filtered moves, thus if two
          sub-formulas have the same filtered moves (e.g., pre-filtering is
          not used), the relations are the same;
        * mas.encoded[agent] gives the relation encoding the equivalence for
          agent;
          it only depends on agent, since it only depends on what he observes;
          THIS RELATION DOES NOT ENCODE THE FACT THAT THE STRATEGIES DO NOT
          CHANGE;
        * mas.encoded["jump"] gives the relation encoding the fact that the
          state is the same but the strategies can change;
          it only depends on the MAS, since all strategies are free and only
          the state is kept identical.
    """
    if not hasattr(mas, "transitions"):
        mas.transitions = {}
    if not hasattr(mas, "encoded"):
        mas.encoded = {}
    
    
    # if mas.encoded["jump"] does not exist,
    # encode it, based on the original state variables of the model
    if "jump" not in mas.encoded:
        jump = jump_relation(mas)
        mas.encoded["jump"] = jump
    
    
    # Encode each agent variables and relations if needed
    
    # for each agent in agents,
    # if (agent, filtered) is not in mas.encoded,
    # compute them
    for agent in agents:
        if (agent, filtered) not in mas.encoded:
            relations = strategy_relations(mas, agents, agent, filtered)
            mas.encoded[(agent, filtered)] = relations
    
    # for each agent in agents,
    # if mas.encoded[agent] is not present,
    # compute it
    for agent in agents:
        if agent not in mas.encoded:
            relation = equivalence_relation(mas, agent)
            mas.encoded[agent] = relation
    
    
    # Encode the transition relations
    if formula not in mas.transitions:
        mas.transitions[formula] = {}
    
    # the "jump" relation is just mas.encoded["jump"]
    if "jump" not in mas.transitions[formula]:
        jump = mas.encoded["jump"]
        trans = BddTrans.from_string(mas.bddEnc.symbTable, str(jump))
        mas.transitions[formula]["jump"] = trans
    
    # the "equiv" relation is based on
    # * the "equiv" relations of each agent in agents
    #   (disjunction, for group knowledge)
    # * the strategies of the agents are the same
    #   (based on the variables of each agent, stored in
    #    mas.encoded[(agent, filtered)])
    if "equiv" not in mas.transitions[formula]:
        equiv = reduce(operator.or_,
                       (mas.encoded[agent] for agent in agents),
                       model.Falseexp())
        equiv = reduce(operator.and_,
                       (mas.encoded[(agent, filtered)][0] for agent in agents),
                       equiv)
        trans = BddTrans.from_string(mas.bddEnc.symbTable, str(equiv))
        mas.transitions[formula]["equiv"] = trans
    
    # the "follow" relation is based on
    # * the original transition relation of the MAS;
    # * the fact that the strategies of the agents are the same,
    #   and that the strategies for agents are followed by these agents,
    #   given by the variables in mas.encoded[(agent, filtered)]
    if "follow" not in mas.transitions[formula]:
        flat = glob.flat_hierarchy()
        follow = reduce(operator.and_,
                        (mas.encoded[(agent, filtered)][0] &
                         mas.encoded[(agent, filtered)][1]
                         for agent in agents),
                        flat.trans)
        trans = BddTrans.from_string(mas.bddEnc.symbTable, str(follow))
        mas.transitions[formula]["follow"] = trans


def jump_relation(mas):
    """
    Return the transition relation corresponding to the "jump" relation of the
    given MAS.
    
    mas -- a multi-agents system.
    
    The returned value is a model-based relation encoding the fact that all
    original state variables of the given MAS stay the same, and other encoded
    variables are free to change.
    """
    # Get the original state variables
    flat = glob.flat_hierarchy()
    symb_table = mas.bddEnc.symbTable
    original_variables = [variable for variable in flat.variables
                          if symb_table.is_state_var(variable)]
    
    # Build the relation expression
    return reduce(operator.and_,
                  (variable.next() == variable
                   for variable in original_variables))


def strategy_relations(mas, agents, agent, filtered):
    """
    Extract and encode the variables representing the strategies of agent in
    filtered, and return
    * the relation encoding the fact that the encoded strategies of agent
      from filtered are kept the same;
    * the relation encoding the fact that the encoded strategies of agent
      are followed;
    
    mas -- a multi-agents system;
    agents -- a set of agents of mas;
    agent -- an agent from agents;
    filtered -- a set of agents-moves.
    
    The returned value is a couple where the first element is the first
    relation, and the second element is the second relation.
    
    The new variables are encoded on a new layer called agent_hash(filtered).
    """
    # Get observable variables and actions for agent
    observables = mas.agents_observed_variables[agent]
    actions = mas.agents_inputvars[agent]
    
    act_cube = mas.bddEnc.cube_for_inputs_vars(actions)
    obs_cube = mas.bddEnc.cube_for_state_vars(observables)
    state_cube = mas.bddEnc.statesCube
    inputs_cube = mas.bddEnc.inputsCube
    other_vars_cube = state_cube - obs_cube
    other_acts_cube = inputs_cube - act_cube
    
    
    new_layer = str(agent) + "_" + str(hash(filtered))
    
    
    # Extract the useful information from filtered
    variables = OrderedDict()
    protocol = (filtered &
                mas.state_constraints &
                mas.inputs_constraints &
                mas.bddEnc.statesInputsMask)
    while protocol.isnot_false():
        si = mas.pick_one_state_inputs(protocol)
        state = mas.pick_one_state(si.forsome(mas.bddEnc.inputsCube))
        
        # Get observation values
        obs_vars = tuple(sorted([(node.Identifier.from_string(var),
                                  node.Expression.from_string(val))
                                 for var, val
                                 in state.get_str_values().items()
                                 if var in observables]))
        
        variables[obs_vars] = list()
        inputs = ((protocol &
                   state.forsome(other_vars_cube)).forsome(
                   mas.bddEnc.statesCube)
                  & mas.bddEnc.statesInputsMask
                  & mas.bddEnc.statesMask
                  & mas.bddEnc.inputsMask
                  & mas.state_constraints
                  & mas.inputs_constraints)
        # Get the possible values for actions
        while inputs.isnot_false():
            i = mas.pick_one_inputs(inputs)
            
            inputs_vars = tuple(sorted(
                          [(node.Identifier.from_string(var),
                            node.Expression.from_string(val))
                           for var, val in i.get_str_values().items()
                           if var in actions]))
            variables[obs_vars].append(inputs_vars)
            
            inputs = inputs - i.forsome(other_acts_cube)
        
        protocol = (protocol - state.forsome(other_vars_cube))
    
    
    # Compute the variables for the strategies of agent in filtered
    new_variables = []
    strategies = list()  # strategies is a list of tuples
                         #   (obs assigment,
                         #    corresponding strategy var,
                         #    corresponding strategy var value,
                         #    actions assigment)
    for obs_values in variables:
        var_name = (new_layer + "_" + assign_to_name(obs_values))
        var_name = node.Identifier.from_string(var_name)
        type_values = []
        for action_values in variables[obs_values]:
            type_value = assign_to_name(action_values)
            type_values.append(type_value)
            
            strategies.append((obs_values,
                               var_name,
                               type_value,
                               action_values))
        
        var_type = mas.bddEnc.symbTable._get_type_from_node(node.Scalar(
                                                                  type_values))
        
        # a triplet (variable name (as Nodes),
        #            variable type (as SymbType),
        #            variable kind (as int))
        new_variables.append((var_name,
                              var_type,
                              mas.bddEnc.symbTable.SYMBOL_STATE_VAR))
    
    new_variables = sorted(new_variables, key=lambda e: str(e[0]))
    strategies = sorted(strategies, key=lambda e: str(e[1]))
    
    
    # Encode them
    symb_table = mas.bddEnc.symbTable
    symb_table.create_layer(new_layer)
    
    # Encode variables
    for var, type_, kind in new_variables:
        if kind is symb_table.SYMBOL_STATE_VAR:
            symb_table.declare_state_var(new_layer, var, type_)
        elif kind is symb_table.SYMBOL_FROZEN_VAR:
            symb_table.declare_frozen_var(new_layer, var, type_)
        elif kind is symb_table.SYMBOL_INPUT_VAR:
            symb_table.declare_input_var(new_layer, var, type_)
    glob.encode_variables_for_layers(layers=[new_layer])
    
    
    # Build the relation
    # strategy stays the same
    strategy_stay = reduce(operator.and_,
                           (var.next() == var for var, _, _ in new_variables))
    
    # actions are followed:
    #   observations are true
    #   & corresponding strategy var is ivar#value##...
    #   ->
    #   ivar = value & ...
    # strategies is a set of tuples
    #   (obs assigment,
    #    corresponding strategy var,
    #    corresponding strategy var value,
    #    actions assigment)
    trans = []
    for move in strategies:
        obs_assign, strat_var, strat_val, act_assign = move
        
        trans_cond = reduce(lambda e, n: e & n,
                            [var == val for var, val in obs_assign],
                            node.Trueexp())
        trans_cond = trans_cond & (strat_var == strat_val)
        
        trans_cons = reduce(lambda e, n: e & n,
                            [var == val for var, val in act_assign],
                            node.Trueexp())
        trans.append(trans_cond.implies(trans_cons))
    followed = reduce(operator.and_, trans)
    
    return strategy_stay, followed


def equivalence_relation(mas, agent):
    """
    Return the (text-based) transition relation corresponding to the
    observations of agent in mas.
    
    mas -- a multi-agents system;
    agent -- an agent of mas.
    
    The returned value is the model-encoded transition relation telling that
    agent's observations are kept the same.
    """
    # compute the relation based on the observable variables of the agent
    trans = reduce(operator.and_,
                   [node.Identifier.from_string(str(obs)) ==
                    node.Identifier.from_string(str(obs)).next()
                    for obs in mas.agents_observed_variables[agent]],
                   node.Trueexp())
    return trans


# ----- eval algorithms -------------------------------------------------------

def _nfair(mas, formula, agents):
    """
    Return the set of states in which the given agents cannot avoid a fair path
    by using the strategy encoded in these states.
    """
    jump = mas.transitions[formula]["jump"]
    equiv = mas.transitions[formula]["equiv"]
    follow = mas.transitions[formula]["follow"]
    
    if not mas.fairness_constraints:
        return BDD.false(mas)
    
    else:
        # nfair =
        # mu Z. \/_fc []_group_follow(nu Y. (Z \/ ~fc) /\ []_group_follow(Y))
        def inner(Z):
            # \/_fc []_group_follow(nu Y. (Z \/ ~fc) /\ []_group_follow(Y))
            res = BDD.false(mas)
            for fc in mas.fairness_constraints:
                fc = fc.forsome(mas.bddEnc.inputsCube)
                nfc = ~fc
                res = res | ~follow.pre(~fixpoint(lambda Y: (Z | nfc) &
                                                            ~follow.pre(~Y),
                                                  BDD.true(mas)))
            return res
        res = fixpoint(inner, BDD.false(mas))
        return res


def eval_cex(mas, formula, agents, states):
    jump = mas.transitions[formula]["jump"]
    equiv = mas.transitions[formula]["equiv"]
    follow = mas.transitions[formula]["follow"]
    reachable = mas.reachable_states
    
    if not mas.fairness_constraints:
        # <group>i X p =
        # <>_group_jump []_group_equiv (reachable => []_group_follow p)
        return jump.pre(~equiv.pre(~(reachable.imply(~follow.pre(~states)))))
    else:
        # <group>i X fair p =
        # <>_group_jump []_group_equiv
        # (reachable => []_group_follow (p \/ ~ifair))
        return jump.pre(~equiv.pre(~(reachable.imply(
                        ~follow.pre(~(states |
                                      _nfair(mas, formula, agents)))))))


def eval_ceu(mas, formula, agents, states_1, states_2):
    jump = mas.transitions[formula]["jump"]
    equiv = mas.transitions[formula]["equiv"]
    follow = mas.transitions[formula]["follow"]
    reachable = mas.reachable_states
    
    if not mas.fairness_constraints:
        # <group>i[p U q] =
        # <>_group_jump []_group_equiv
        # (reachable => mu Z . q \/ (p /\ []_group_follow Z))
        return jump.pre(~equiv.pre(~(reachable.imply(
                        fixpoint(lambda Z: states_2 |
                                           (states_1 & ~follow.pre(~Z)),
                                 BDD.false(mas))))))
    else:
        nfair = _nfair(mas, formula, agents)
        # <group>i fair [p U q] =
        # <>_group_jump []_group_equiv
        # reachable =>
        # mu Z. (p \/ q \/ ~ifair) /\
        #       (q \/_fc []_group_follow
        #                (nu Y. (Z \/ ~fc) /\
        #                              (p \/ q \/ ~ifair) /\
        #                              (q \/ []_group_follow(Y))))
        def inner(Z):
            # (p \/ q \/ ~ifair) /\
            # (q \/_fc []_group_follow
            #          (nu Y. (Z \/ ~fc) /\
            #                        (p \/ q \/ ~ifair) /\
            #                        (q \/ []_group_follow(Y))))
            res = BDD.false(mas)
            for fc in mas.fairness_constraints:
                fc = fc.forsome(mas.bddEnc.inputsCube)
                nfc = ~fc
                # []_group_follow (nu Y. (Z \/ ~fc) /\
                #                        (p \/ q \/ ~ifair) /\
                #                        (q \/ []_group_follow(Y)))
                res = res | ~follow.pre(
                      ~(fixpoint(lambda Y: ((Z | nfc) &
                                            (states_1 | states_2 | nfair) &
                                            (states_2 | ~follow.pre(~Y))),
                                 BDD.true(mas))))
            return ((states_1 | states_2 | nfair) & (states_2 | res))
        return jump.pre(~equiv.pre(~(reachable.imply(
                        fixpoint(inner, BDD.false(mas))))))


def eval_cew(mas, formula, agents, states_1, states_2):
    jump = mas.transitions[formula]["jump"]
    equiv = mas.transitions[formula]["equiv"]
    follow = mas.transitions[formula]["follow"]
    reachable = mas.reachable_states
    
    if not mas.fairness_constraints:
        # <group>i[p W q] = 
        # <>_group_jump []_group_equiv
        # (reachable => nu Z . q \/ (p /\ []_group_follow Z))
        return jump.pre(~equiv.pre(~(reachable.imply(
                        fixpoint(lambda Z: states_2 |
                                           (states_1 & ~follow.pre(~Z)),
                                 BDD.true(mas))))))
    else:
        nfair = _nfair(mas, formula, agents)
        # <group>i fair [p W q] =
        # <>_group_jump []_group_equiv
        # reachable =>
        # nu Z. (p \/ q \/ ~ifair) /\ (q \/ []_group_follow(Z))
        res = jump.pre(~equiv.pre(~(reachable.imply(
                       fixpoint(lambda Z: ((states_1 | states_2 | nfair) &
                                           (states_2 | ~follow.pre(~Z))),
                                BDD.true(mas))))))
        return res


# ----- CTL algorithms --------------------------------------------------------

def _fair(mas):
    if not mas.fairness_constraints:
        return BDD.true(mas)
    else:
        run = mas.trans
        # fair = nu Z. /\_fc pre(mu Y. (Z /\ fc) \/ pre(Y))
        def inner(Z):
            res = BDD.true(mas)
            for fc in mas.fairness_constraints:
                fc = fc.forsome(mas.bddEnc.inputsCube)
                res = res & run.pre(fixpoint(lambda Y: (Z & fc) | run.pre(Y),
                                             BDD.false(mas)))
            return res
        return fixpoint(inner, BDD.true(mas))


def ex(mas, states):
    run = mas.trans
    return run.pre(states & _fair(mas))

def eu(mas, states_1, states_2):
    run = mas.trans
    return fixpoint(lambda Y : (states_2 & _fair(mas)) |
                               (states_1 & run.pre(Y)),
                    BDD.false(mas))


def ew(mas, states_1, states_2):
    run = mas.trans
    if not mas.fairness_constraints:
        return fixpoint(lambda Z: states_2 | (states_1 & run.pre(Z)),
                        BDD.true(mas))
    else:
        def inner(Z):
            res = BDD.true(mas)
            for fc in mas.fairness_constraints:
                fc = fc.forsome(mas.bddEnc.inputsCube)
                res = res & run.pre(fixpoint(lambda Y: (states_2 & _fair(mas))
                                                       |
                                                       (Z & fc) |
                                                       (states_1 & run.pre(Y)),
                                             BDD.false(mas)))
            return (res & states_1) | (states_2 & _fair(mas))
        return fixpoint(inner, BDD.true(mas))


# ----- caching ---------------------------------------------------------------

__evalATLK_cache = {}
__orig_evalATLK = evalATLK
def __cached_evalATLK(mas, formula, pre_filtering=False):
    if (mas, formula) not in __evalATLK_cache:
        sat = __orig_evalATLK(mas, formula, pre_filtering=pre_filtering)
        __evalATLK_cache[(mas, formula)] = sat
    return __evalATLK_cache[(mas, formula)]
evalATLK = __cached_evalATLK



from .naive import evalATLK as eval_naive
from .partial import evalATLK as eval_partial
from .early import evalATLK as eval_early
from .symbolic import evalATLK as eval_symbolic
from .backward import evalATLK as eval_backward

__implementations = {"naive": eval_naive,
                     "partial": eval_partial,
                     "early": eval_early,
                     "symbolic": eval_symbolic,
                     "backward": eval_backward}

def check(mas, formula, implementation="naive", pre_filtering=False):
    """
    Return whether the system satisfies the ATLK formula.
    
    mas -- a multi-agents system;
    formula -- the specification, as a string;
    implem -- the particular implementation to use for strategic operators
              evaluation:
              * "naive" the naive implementation;
              * "partial" a version based on partial strategies;
              * "early" a version based on early evaluated partial strategies;
              * "symbolic" a version based on a fully symbolic approach.
    """
    if implementation in __implementations:
        sat = __implementations[implementation](mas,
                                                formula,
                                                pre_filtering=pre_filtering)
    else:
        raise Exception("check: unknown implementation: " + implementation)
    return (~sat & mas.bddEnc.statesInputsMask & mas.init).is_false()



"""
Graphs inspector.

The graph inspector is a tkinter-based GUI to inspect and manipulate graphs.
"""

import tkinter as tk
import tkinter.ttk as ttk
from copy import deepcopy
from itertools import count
from collections import OrderedDict, defaultdict, Mapping, Iterable
from tkCanvasGraph import CanvasFrame, Vertex, Edge
from tkCanvasGraph.mouse import Mouse
from tkCanvasGraph.util import ObservableSet


class ObservableElement(object):
    """
    An observable element. Notifies registered observers when value changes.
    """
    def __init__(self, value=None):
        self._value = value
        self._observers = []
    
    @property
    def value(self):
        """
        The value of this element.
        """
        return self._value
    
    @value.setter
    def value(self, new_value):
        self._value = new_value
        self.notify()
    
    def notify(self):
        """
        Notify all registered observers.
        """
        for observer in self._observers:
            observer.update(self)
    
    def register(self, observer):
        """
        Register the given observer. The observer will be notified when this
        element is updated.
        
        observer -- the observer to notify; must have an update(element)
                    method.
        """
        self._observers.append(observer)

    def unregister(self, observer):
        """
        Unregister the given observer. It will not be notified any more.

        observer -- the observer to remove.
        """
        try:
            self._observers.remove(observer)
        except ValueError:
            pass


class KeysDictionary(OrderedDict):
    """
    A keys dictionary is a dictionary with key "_all_" leading to a boolean
    value, and any other key leads to keys dictionaries.
    """
    
    def __init__(self, *args, **kwargs):
        """
        Create a new keys dictionary from the given data.
        
        The given data must be compliant with the definition of keys
        dictionary.
        """
        super().__init__(*args, **kwargs)
    
    
    def merge(self, other):
        """
        Merge this keys dictionary with other
        
        second -- another keys dictionary.
        
        The result is a new keys dictionary with the keys of first and second,
        and where the values are merged, that is, if the key is _all_, the
        value is the one of second, if it exists, otherwise the value of first,
        and where the values of the other keys are recursively merged.
        """
        first = deepcopy(self)
        for key, value in other.items():
            if key in first and isinstance(value, KeysDictionary):
                first[key] = first[key].merge(value)
            else: # key is not in first or key is _all_
                first[key] = value
        return first

    def some(self, value):
        """
        Return whether somes keys of this keys dictionary have value or not
        (recursively).
        
        value -- a boolean.
        """
        for k, v in self.items():
            if v == value:
                return True
            if isinstance(v, KeysDictionary) and v.some(value):
                return True
        return False

    def set_all(self, value):
        """
        Recursively set all values of this keys dictionary to value.
        
        value -- a boolean.
        """
        self["_all_"] = value
        for k, v in self.items():
            if isinstance(v, KeysDictionary):
                v.set_all(value)

    def hide_empty(self):
        """
        For all keys where all subkeys are not visible, set the keys as
        non-visible, recursively.
        """
        all_false = True
        for key, value in self.items():
            if key != "_all_":
                value.hide_empty()
                if value.some(True):
                    all_false = False
        if all_false and len(self) > 1:
            self["_all_"] = False

    def show_nonempty(self):
        """
        For all keys where some subkeys are visible, set the keys as visible,
        recursively.
        """
        some_true = False
        for key, value in self.items():
            if key != "_all_":
                value.show_nonempty()
                if value.some(True):
                    some_true = True
        if some_true:
            self["_all_"] = True

    def harmonize(self):
        """
        For all keys where all subkeys are not visible, set the keys as
        non-visible, recursively.
        For all keys where some subkeys are visible, set the keys as visible,
        recursively.
        """
        self.hide_empty()
        self.show_nonempty()


class GraphVertex(Vertex):
    """
    A vertex representing a graph node.
    """
    def __init__(self, inspector, node):
        super().__init__(inspector.canvas)
        self.node = node
        
        # The values to display as label
        # if node contains a _label_ key, use it as the values:
        #   if _label_ is callable, call it and analyse the result further
        #   if _label_ (or the result of the call) is a string,
        #   use it directly as the label
        #   otherwise, use the value of _label_ (or the result of the call)
        #   as it is
        # otherwise, the node itself (a dictionary) is used as the values

        if callable(node._label_):
            values = node._label_(inspector)
        else:
            values = node._label_
        if values is None:
            values = node
        self._label_values = values
        
        # The view of the vertex
        # if node contains a _view_ key, use it as the values:
        #   if _view_ is callable, call it and analyse the result further
        #   then use the value of _view_ (or the result of the call) as it is
        # otherwise, the node itself (a dictionary) is used as the values
        if callable(node._view_):
            view = node._view_(inspector)
        else:
            view = node._view_
        if view is None:
            view = node
        self._view = view
        
        # The menu of the vertex
        # if node contains a _menu_ key, use it as the menu:
        #   if _menu_ is callable, call it and analyse the result further
        #   then use the value of _menu_ (or the result of the call) as it is
        # otherwise, no menu is considered
        if callable(node._menu_):
            menu = node._menu_(inspector)
        else:
            menu = node._menu_
        self._menu = menu
    
    @property
    def data(self):
        return self.node
    
    @property
    def label_values(self):
        return self._label_values
    
    @property
    def view(self):
        return self._view
    
    @property
    def menu(self):
        return self._menu


class GraphEdge(Edge):
    """
    An edge representing a graph edge.
    """
    def __init__(self, inspector, origin, end, edge):
        super().__init__(inspector.canvas, origin, end)
        self.edge = edge
        
        # The values to display as label
        # if edge contains a _label_ key, use it as the values:
        #   if _label_ is callable, call it and analyse the result further
        #   if _label_ (or the result of the call) is a string,
        #   use it directly as the label
        #   otherwise, use the value of _label_ (or the result of the call)
        #   as it is
        # otherwise, the edge itself (a dictionary) is used as the values
        if callable(edge._label_):
            values = edge._label_(inspector)
        else:
            values = edge._label_
        if values is None:
            values = edge
        self._label_values = values
        
        # The view of the edge
        # if edge contains a _view_ key, use it as the values:
        #   if _view_ is callable, call it and analyse the result further
        #   then use the value of _view_ (or the result of the call) as it is
        # otherwise, the edge itself (a dictionary) is used as the values
        if callable(edge._view_):
            view = edge._view_(inspector)
        else:
            view = edge._view_
        if view is None:
            view = edge
        self._view = view
        
        # The menu of the edge
        # if edge contains a _menu_ key, use it as the menu:
        #   if _menu_ is callable, call it and analyse the result further
        #   then use the value of _menu_ (or the result of the call) as it is
        # otherwise, no menu is considered
        if callable(edge._menu_):
            menu = edge._menu_(inspector)
        else:
            menu = edge._menu_
        self._menu = menu
    
    @property
    def data(self):
        return self.edge
    
    @property
    def label_values(self):
        return self._label_values
    
    @property
    def view(self):
        return self._view
    
    @property
    def menu(self):
        return self._menu


class GraphFrame(CanvasFrame):
    """
    A frame displaying a graph.
    """
    pass


def _get_path(elements):
    """
    Return the path represented by elements, or None if elements does
    not define a path.
    
    elements -- a set of elements.
    
    The returned path is a list where vertices alternate with edges.
    The elements might define a loop; in this case, the first vertex is chosen
    arbitrarily.
    The elements might define a lasso; in this case, the first unrolling of the
    loop is exposed.
    """
    edges = {e for e in elements if isinstance(e, GraphEdge)}
    vertices = {e for e in elements if isinstance(e, GraphVertex)}
    graph = {v: {e for e in edges if e.origin == v}
             for v in vertices}
    
    # Get source
    sources = {v for v in graph
               if len({e for e in edges if e.end == v}) <= 0}
    if len(sources) > 1:
        # More than one source, it does not define a single path
        return None
    elif not sources:
        # No source, pick any vertex
        source = next(iter(vertices))
    else:
        # One source
        source = next(iter(sources))
    
    path = [source]
    while True:
        outgoings = graph[path[-1]]
        if len(outgoings) > 1:
            # More than one outgoing edge, this is not a single path
            return None
        elif not outgoings:
            # No more nodes, end of path
            break
        
        outgoing = next(iter(outgoings))
        
        if outgoing.end in path:
            # we reach a loop
            path.append(outgoing)
            path.append(outgoing.end)
            break
        
        else:
            path.append(outgoing)
            path.append(outgoing.end)
    
    return path


class PathInspectorFrame(tk.Frame):
    """
    A frame to inspect a path in a graph.
    """
    def __init__(self, inspected, parent, **config):
        """
        Create a new inspector for the given observable set of elements.
        
        inspected -- an observable set of elements.
        """
        super(PathInspectorFrame, self).__init__(parent, **config)
        self.content = tk.Frame(master=self)
        self.content.pack(fill=tk.BOTH, expand=True)
        # Register to the explanations set
        self.inspected = inspected
        inspected.register(self)
        
        self.update(inspected)
    
    def _remove_content(self):
            """
            Remove the content of this frame.
            """
            self.content.destroy()
            self.content = tk.Frame(self)
            self.content.pack(fill=tk.BOTH, expand=True)

    def update(self, inspected):
        """
        Update this frame to show inspected path.
        """
        self._remove_content()
        
        if len(inspected) <= 0:
            label = tk.Label(self.content, text="No inspected path.")
            label.pack()
        else:
            path = _get_path(inspected)
            if path is None:
                label = tk.Label(self.content,
                                 text=
                                 "Highlighted elements do not define a path.")
                label.pack()
            else:
                # Treeview for paths
                
                vertices = [path[0]] + path[2::2]
                edges = path[1::2]
                
                # Create the treeview
                self.content.grid_rowconfigure(0, weight=1)
                self.content.grid_columnconfigure(0, weight=1)
                treeview = ttk.Treeview(self.content,
                                        columns=["c" + str(i) for i in
                                                 range(len(vertices))])
                # Set columns widths
                treeview.column("#0", width=50)
                for col in range(len(vertices)):
                    treeview.column("c" + str(col), width=50)
                
                treeview.heading("#0", text="Key")
                for col in range(len(vertices)):
                    treeview.heading("c" + str(col),
                                     text="Value " + str(col + 1))
                
                # Add scrollbars
                ysb = tk.Scrollbar(self.content, orient=tk.VERTICAL,
                                   command=treeview.yview)
                xsb = tk.Scrollbar(self.content, orient=tk.HORIZONTAL,
                                    command=treeview.xview)
                treeview.configure(yscrollcommand=ysb.set,
                                   xscrollcommand=xsb.set)
                
                treeview.grid(row=0, column=0, sticky=tk.W+tk.E+tk.N+tk.S)
                ysb.grid(row=0, column=1, sticky=tk.NS)
                xsb.grid(row=1, column=0, sticky=tk.EW)
                
                # Get views and keys
                vertices_views = [vertex.view for vertex in vertices]
                edges_views = [edge.view for edge in edges]
                vertices_keys = get_all_keys(vertices_views)
                edges_keys = get_all_keys(edges_views)
                
                # Gather data
                # Get a dictionary with one list of lists in _all_
                # and other keys contain similar dictionaries
                # the ith list of _all_ contains the list of values to display
                # for the ith element of the path
                # the other keys lead to similar dictionaries containing
                # the values in the different elements of the path, under
                # the corresponding key
                def gather_data(view, view_id, nbviews, gatherer=None):
                    if gatherer is None:
                        gatherer = OrderedDict()
                    
                    if isinstance(view, Mapping):
                        for key, value in view.items():
                            if not str(key).startswith("_"):
                                if key not in gatherer:
                                    gatherer[key] = OrderedDict()
                                gather_data(value,
                                            view_id,
                                            nbviews,
                                            gatherer[key])
                    
                    elif (isinstance(view, Iterable) and
                          not isinstance(view, str)):
                        for element in view:
                            gather_data(element, view_id, nbviews, gatherer)
                    
                    else:
                        if "_all_" not in gatherer:
                            gatherer["_all_"] = [[] for _ in range(nbviews)]
                        gatherer["_all_"][view_id].append(view)
                    
                    if "_all_" in gatherer:
                        gatherer.move_to_end("_all_", last=False)
                    
                    return gatherer
                
                vertices_data = None
                for i, view in enumerate(vertices_views):
                    vertices_data = gather_data(view,
                                                i,
                                                len(vertices_views),
                                                gatherer=vertices_data)
                
                edges_data = None
                edges_views = [{}] + edges_views
                for i, view in enumerate(edges_views):
                    edges_data = gather_data(view,
                                             i,
                                             len(edges_views),
                                             gatherer=edges_data)
                
                # Add data to treeview
                curid = count(1)
                def add_data(treeview, data, parentid):
                    # data is a list of lists
                    while data != [[]] * len(data):
                        values = []
                        for view in data:
                            if len(view):
                                values.append(view.pop(0))
                            else:
                                values.append("")
                        iid = parentid + str(next(curid))
                        treeview.insert(parentid,
                                        "end",
                                        iid=iid,
                                        value=values)
                
                def add(treeview, data, parentid):
                    # data is a dictionary
                    if "_all_" in data:
                        add_data(treeview, data["_all_"], parentid)
                    for key, values in data.items():
                        if not str(key).startswith("_"):
                            iid = parentid + key
                            
                            if (len(values) <= 1 and
                                "_all_" in values and
                                all(len(view) <= 1
                                    for view in values["_all_"])):
                                vv = [(str(view.pop(0))
                                       if len(view)
                                       else "")
                                      for view in values["_all_"]]
                                treeview.insert(parentid,
                                                "end",
                                                iid=iid,
                                                text=str(key),
                                                open=True,
                                                values=vv)
                            
                            else:
                                treeview.insert(parentid,
                                                "end",
                                                iid=iid,
                                                text=str(key),
                                                open=True)
                                add(treeview, values, iid)
                
                add(treeview, edges_data, "")
                treeview.insert("", "end")
                add(treeview, vertices_data, "")


class InspectorModifyingMouse(Mouse):
    """
    A mouse that adds and removes inspected elements
    """
    
    def __init__(self, inspected, edges):
        super(InspectorModifyingMouse, self).__init__()
        self.inspected = inspected
        self.edges = edges
    
    def pressed(self, event):
        """
        React to mouse button pressed.
        """
        element = event.element
        if element is not None:
            if element not in self.inspected:
                if isinstance(element, GraphEdge):
                    self.inspected.add(element.origin)
                    self.inspected.add(element.end)
                self.inspected.add(element)
            else:
                if isinstance(element, GraphVertex):
                    for edge in self.edges:
                        if (edge.origin == element or
                            edge.end == element):
                            self.inspected.discard(edge)
                self.inspected.discard(element)
            
            event.canvas.refresh()
            return False
        return True

class PathInspectingMouse(Mouse):
    """
    A mouse that inspect one path.
    """
    
    def __init__(self, inspected, vertices, edges):
        """
        inspected -- the observable set of inspected elements;
        vertices -- the set of vertices of the canvas;
        edges the set of edges of the canvas.
        """
        super(PathInspectingMouse, self).__init__()
        self.inspected = inspected
        self.vertices = vertices
        self.edges = edges
        
        self.arrow = None
        self.start_arrow = None
        
        self.is_pressed = False
        self.start = None
    
    def pressed(self, event):
        """Mouse button pressed."""
        self.start = event.element
        
        if event.element is not None:
            self.start_arrow = event.position
        return True
    
    def moved(self, event):
        """Mouse moved."""
        def delta(a, b):
            if a > b:
                return a - b
            else:
                return b - a
        
        if self.start_arrow is not None:
            delta_x = delta(self.start_arrow[0], event.position[0])
            delta_y = delta(self.start_arrow[1], event.position[1])
            if delta_x > 2 or delta_y > 2:
                if self.arrow is None:
                    self.arrow = event.canvas.create_line(self.start_arrow +
                                                          self.start_arrow,
                                                          fill="gray",
                                                          arrow="last",
                                                          width=3)
                end_position = ((event.position[0] + 2)
                                if event.position[0] < self.start_arrow[0]
                                else (event.position[0] - 2),
                                (event.position[1] + 2)
                                if event.position[1] < self.start_arrow[1]
                                else (event.position[1] - 2))
            
                coords = self.start_arrow + end_position
                event.canvas.coords(self.arrow, *coords)
        return True
    
    def released(self, event):
        """Mouse button released."""
        self.inspected.clear()
        element = event.element
        
        # If on start, select start (and ends if an edge)
        if element == self.start:
            if isinstance(element, Vertex):
                self.inspected.add(element)
            elif isinstance(element, Edge):
                self.inspected.add(element.origin)
                self.inspected.add(element.end)
                self.inspected.add(element)
        
        # If on another than start, and both are vertices, select path
        elif (isinstance(self.start, Vertex) and
              isinstance(element, Vertex)):
              # Get shortest path
              path = self.shortest_path(self.start, element)
              # Highlight the shortest path
              if path is not None:
                  self.inspected.clear()
                  self.inspected.update(path)
        
        # Else, inspect nothing
        
        # Remove arrow
        if self.arrow is not None:
            event.canvas.delete(self.arrow)
            self.arrow = None
        self.start_arrow = None
        
        event.canvas.refresh()
        
        self.start = None
        return True
    
    def shortest_path(self, start, end):
        """
        Return the shortest path between start and end.
        
        Return None if no such path exists.
        """
        graph = {v: {e for e in self.edges if e.origin == v}
                 for v in self.vertices}
        
        pred = {start: None}
        to_expand = [start]
        
        # BFS
        while len(to_expand) > 0 and end not in pred:
            current = to_expand[0]
            to_expand = to_expand[1:]
            
            for edge in graph[current]:
                if edge.end not in pred:
                    to_expand.append(edge.end)
                    pred[edge.end] = current
        
        if end not in pred:
            return None
        
        # Build path
        path = [end]
        while path[-1] != start:
            o, e = pred[path[-1]], path[-1]
            edge = next(iter({ed for ed in graph[o] if ed.end == e}))
            path.append(edge)
            path.append(o)
        path.reverse()
        
        return path


class ElementInspectorFrame(tk.Frame):
    """
    A frame to inspect a particular element.
    """
    
    def __init__(self, inspected, parent, **config):
        """
        Create a new inspector for the given observable element.
        
        inspected -- an observable element.
        """
        # Create the frame
        super(ElementInspectorFrame, self).__init__(parent, **config)
        self.content = tk.Frame(master=self)
        self.content.pack(fill=tk.BOTH, expand=True)
        # Register to the explanations set
        self.inspected = inspected
        inspected.register(self)
        
        self.update(inspected)
    
    def _remove_content(self):
        """
        Remove the content of this frame.
        """
        self.content.destroy()
        self.content = tk.Frame(self)
        self.content.pack(fill=tk.BOTH, expand=True)
    
    def update(self, inspected):
        """
        Update this frame to show inspected element.
        """
        self._remove_content()
        
        def has_keys(view):
            if isinstance(view, Mapping):
                return True
            elif isinstance(view, Iterable) and not isinstance(view, str):
                for element in view:
                    if has_keys(element):
                        return True
                return False
            else:
                return False
        
        if not inspected.value:
            label = tk.Label(self.content, text="No inspected element.")
            label.pack()
        else:
            element = inspected.value
            view = element.view
            
            self.content.grid_rowconfigure(0, weight=1)
            self.content.grid_columnconfigure(0, weight=1)
            
            if has_keys(view):
                curid = count(1)
                def add(treeview, element, parentid):
                    if isinstance(element, Mapping):
                        for key, value in element.items():
                            if not str(key).startswith("_"):
                                iid = (parentid +
                                       "." +
                                       str(key) +
                                       str(next(curid)))
                                if (isinstance(value, Mapping) or
                                    (isinstance(value, Iterable) and
                                     not isinstance(value, str))):
                                    treeview.insert(parentid,
                                                    "end",
                                                    iid=iid,
                                                    text=str(key),
                                                    open=True)
                                    add(treeview, value, iid)
                                else:
                                    treeview.insert(parentid,
                                                    "end",
                                                    iid=iid,
                                                    text=str(key),
                                                    value=(str(value),))
                    elif (isinstance(element, Iterable) and
                          not isinstance(element, str)):
                        for value in element:
                            add(treeview, value, parentid)
                    else:
                        iid = parentid + "." + str(next(curid))
                        treeview.insert(parentid,
                                        "end",
                                        iid=iid,
                                        value=(str(element),))
                
                treeview = ttk.Treeview(self.content, columns=["value"])
                
                treeview.heading("#0", text="Key")
                treeview.heading("value", text="Value")
                
                treeview.column("#0", width=100)
                treeview.column("value", width=100)
                
                add(treeview, view, "")
            
            else: # not has_keys(view)
                def add(treeview, element):
                    if (isinstance(element, Iterable) and
                        not isinstance(element, str)):
                        for value in element:
                            add(treeview, value)
                    else:
                        treeview.insert("", "end",
                                        iid=str(element),
                                        text=str(element),
                                        values=())
                
                treeview = ttk.Treeview(self.content, columns=[])
                treeview.column("#0", width=100)
                
                add(treeview, view)
            
            ysb = tk.Scrollbar(self.content, orient=tk.VERTICAL,
                               command=treeview.yview)
            xsb = tk.Scrollbar(self.content, orient=tk.HORIZONTAL,
                                command=treeview.xview)
            treeview.configure(yscrollcommand=ysb.set,
                               xscrollcommand=xsb.set)
            
            treeview.grid(row=0, column=0, sticky=tk.W+tk.E+tk.N+tk.S)
            ysb.grid(row=0, column=1, sticky=tk.NS)
            xsb.grid(row=1, column=0, sticky=tk.EW)


class ElementInspectorMouse(Mouse):
        """
        A mouse that inspects one element.
        """
        
        def __init__(self, inspected):
            super(ElementInspectorMouse, self).__init__()
            self.inspected = inspected
        
        def pressed(self, event):
            """
            React to mouse button pressed.
            """
            self.inspected.value = event.element
            event.canvas.refresh()
            return False

class KeysChooserMouse(Mouse):
        """
        A mouse that choose the keys to display.
        """
        
        def __init__(self, inspector):
            super(KeysChooserMouse, self).__init__()
            self.inspector = inspector
        
        def pressed(self, event):
            """
            React to mouse button pressed.
            """
            
            def menu_for_keys(keys, parent, top_keys):
                """
                Return the menu for the given keys dictionary.
                
                keys -- a keys dictionary.
                parent -- the parent of the created menu.
                """
                # List the keys, except _all_
                # If the corresponding value is only {"_all_": val},
                # make it a checkable menu item
                # otherwise make it a cascading menu with the menu for the
                # keys
                menu = tk.Menu(parent)
                for key, value in keys.items():
                    if key != "_all_":
                        if len(value) == 1: # Only "_all_"
                            bvar = tk.BooleanVar()
                            bvar.set(value["_all_"])
                            def callback(*args,
                                         bvar=bvar,
                                         value=value):
                                value["_all_"] = bool(bvar.get())
                                top_keys.harmonize()
                                event.canvas.refresh()
                            bvar.trace("w", callback)
                            menu.add_checkbutton(label=key,
                                                 onvalue=True,
                                                 offvalue=False,
                                                 variable=bvar)
                        
                        else: # Some sub keys
                            submenu = menu_for_keys(value, menu, top_keys)
                            menu.add_cascade(menu=submenu, label=key)
                
                # Special treatment if top-level is only _all_
                if len(keys) == 1:
                    def callback(*args):
                        keys["_all_"] = not keys["_all_"]
                        top_keys.harmonize()
                        event.canvas.refresh()
                    label = "Hide" if keys["_all_"] else "Show"
                    menu.add_command(label=label, command=callback)
                
                menu.add_separator()
                
                # If there is more than one key
                # If some are displayed, add Hide all
                # If some are hidden, add Show all
                if len(keys) > 2:
                    def hide_all():
                        keys.set_all(False)
                        event.canvas.refresh()
                    def show_all():
                        keys.set_all(True)
                        event.canvas.refresh()
                    if keys.some(True):
                        menu.add_command(label="Hide all",
                                         command=hide_all)
                    if keys.some(False):
                        menu.add_command(label="Show all",
                                         command=show_all)
                
                return menu
            
            if event.element is None:
                menu = tk.Menu(event.canvas)
                display_menu = tk.Menu(menu)
                
                nodes_menu = menu_for_keys(self.inspector.nodes_keys,
                                           display_menu,
                                           self.inspector.nodes_keys)
                
                edges_menu = menu_for_keys(self.inspector.edges_keys,
                                           display_menu,
                                           self.inspector.edges_keys)
                
                display_menu.add_cascade(menu=nodes_menu, label="Vertices")
                display_menu.add_cascade(menu=edges_menu, label="Edges")
                menu.add_cascade(menu=display_menu, label="Display")
                
                menu.post(*event.absolute)
                return False
            
            else:
                # No handling for single element menu
                return True


class CustomMenuMouse(Mouse):
    """
    Display the menu on vertices and edges if present.
    """
    
    def pressed(self, event):
        """
        React to mouse button pressed.
        """
        if event.element is not None:
            menu = event.element.menu
            if menu is not None:
                menu.post(*event.absolute)
                return False
            else:
                return True
        else:
            return True


def label(element, displayed_keys):
    """
    Return the label of the element.
    
    element -- the element to get the label from.
    displayed_keys -- the keys to display.
    """
    
    def indent(text, size=4, chr=" "):
        """
        Indent the given text with size occurrences of chr.
        """
        return "\n".join(size * chr + line for line in text.split("\n"))
    
    values = element
    if isinstance(values, str):
        if displayed_keys["_all_"]:
            return values
        else:
            return ""
    
    # Build the string (values is either an iterable or a mapping)
    #
    # mappings are displayed as:
    #   key = value, where value is recursively built with the same process
    # only keys that do not start with _ are used
    #
    # iterables (other than strings) are displayed one value per line
    # (where the values are recursively built with the same process)
    #
    # if a value is more than one line long (contains a newline),
    # it is indented with 4 spaces and printed on a new line
    
    # To define what should be displayed,
    # displayed_keys is a dictionary
    #   with key _all_ telling whether the whole element should be displayed
    #   or not
    #   and with the keys of the dictionary telling whether they should be
    #   displayed invididually.
    
    def build_iterable(values, displayed_keys=None):
        if displayed_keys is not None and displayed_keys["_all_"]:
            built = []
            for element in values:
                if isinstance(element, Mapping):
                    built_element = build_mapping(element, displayed_keys)
                elif (isinstance(element, Iterable) and
                      not isinstance(element, str)):
                    built_element = build_iterable(element, displayed_keys)
                else:
                    built_element = str(element)
                
                built.append(built_element)
            
            return "\n".join(built)
        else:
            return ""
        
    def build_mapping(values, displayed_keys=None):
        if displayed_keys is not None and displayed_keys["_all_"]:
            built = []
            for key, element in values.items():
                if (not str(key).startswith("_") and
                    displayed_keys[key]["_all_"]):
                    
                    if isinstance(element, Mapping):
                        built_element = build_mapping(element,
                                                      displayed_keys=
                                                      displayed_keys[key])
                    elif (isinstance(element, Iterable) and
                          not isinstance(element, str)):
                        built_element = build_iterable(element,
                                                       displayed_keys=
                                                       displayed_keys[key])
                    else:
                        built_element = str(element)
                    
                    if built_element.find("\n") >= 0:
                        built.append(str(key) +
                                     " =\n" +
                                     indent(built_element))
                    else:
                        built.append(str(key) +
                                     " = " +
                                     built_element)
            
            return "\n".join(built)
        else:
            return ""
    
    if isinstance(values, Mapping):
        return build_mapping(values, displayed_keys=displayed_keys)
    else: # isinstance(values, Iterable)
        return build_iterable(values, displayed_keys=displayed_keys)


def element_keys(element, follow=None):
    """
    Return the keys dictionary of element.
    
    element -- any data structure.
    follow -- a keys dictionary from which the values will be extracted
              for the keys appearing in element.
    
    If element is a Mapping, returns {"_all_": True} augmented with the
    keys dictionary for the other keys, otherwise return {"_all_": True}.
    If follow is not None, the boolean values are retrieved from follow
    for the keys in common.
    
    """
    if follow is None:
        follow = KeysDictionary()
    
    if "_all_" in follow:
        keys = KeysDictionary(_all_=follow["_all_"])
    else:
        keys = KeysDictionary(_all_=True)
    
    if isinstance(element, Mapping):
        for key, value in element.items():
            if not str(key).startswith("_"):
                if key in follow:
                    keys[key] = element_keys(value, follow=follow[key])
                else:
                    keys[key] = element_keys(value)
    
    elif isinstance(element, Iterable) and not isinstance(element, str):
        for value in element:
            keys = keys.merge(element_keys(value, follow=follow))
    
    return keys


def get_all_keys(elements, follow=None):
    """
    Return the merged keys dictionary of all elements, in the given other.
    
    elements -- an iterable of data structures.
    follow -- a keys dictionary from which the values will be extracted
              for the keys appearing in element.
    
    Note: the result contains at least {"_all_": True}, even if the iterable is
    empty.
    If follow is not None, the boolean values are retrieved from follow
    for the keys in common.
    """
    
    if follow is None:
        follow = KeysDictionary()
    
    if "_all_" in follow:
        keys = KeysDictionary(_all_=follow["_all_"])
    else:
        keys = KeysDictionary(_all_=True)
    
    for element in elements:
        keys = keys.merge(element_keys(element, follow=follow))
    
    return keys


class InspectorFrame(tk.Frame):
    """
    An inspector frame. Inspect a given graph.
    """
    def __init__(self, graph, parent, data=None, **config):
        """
        Create a new inspector for graph.
        The new inspector frame gets the given parent as parent.
        
        graph -- a relational graph.
        parent -- a tkinter widget.
        config -- a possible frame configuration.
        data -- any additional data to store in the inspector.
        """
        super().__init__(parent, **config)
        
        self.data = data
        
        self._graph = graph
        
        self._inspected_element = ObservableElement()
        self._inspected_path = ObservableSet()
        
        # Populate the window:
        #   One paned window with:
        #       a graph frame on the left
        #       a ttk notebook on the right with:
        #           an Inspect tab
        vertical_split = ttk.PanedWindow(parent, orient=tk.VERTICAL)
        vertical_split.pack(fill=tk.BOTH, expand=True)

        horizontal_split = ttk.PanedWindow(vertical_split,
                                           orient=tk.HORIZONTAL)
        horizontal_split.pack(fill=tk.BOTH, expand=True)
        vertical_split.add(horizontal_split, weight=1)

        # Graph frame
        graph_frame = GraphFrame(horizontal_split)
        self._graph_frame = graph_frame
        graph_frame.pack(fill=tk.BOTH, expand=True)
        horizontal_split.add(graph_frame, weight=1)

        # Element inspector frame
        element_inspector_frame = tk.Frame(horizontal_split)
        element_inspector_title = tk.Label(element_inspector_frame,
                                           text="Element inspector",
                                           font="-weight bold")
        element_inspector_title.pack(fill=tk.NONE, expand=False, anchor=tk.NW)

        element_inspector = ElementInspectorFrame(self._inspected_element,
                                                  element_inspector_frame)
        element_inspector.pack(fill=tk.BOTH, expand=True)
        element_inspector_frame.pack(fill=tk.BOTH, expand=True)
        horizontal_split.add(element_inspector_frame)

        # Path inspector frame
        path_inspector_frame = tk.Frame(vertical_split)
        path_inspector_title = tk.Label(path_inspector_frame,
                                        text="Path inspector",
                                        font="-weight bold")
        path_inspector_title.pack(fill=tk.NONE, expand=False, anchor=tk.NW)
        path_inspector = PathInspectorFrame(self._inspected_path,
                                            path_inspector_frame)
        path_inspector.pack(fill=tk.BOTH, expand=True)
        path_inspector_frame.pack(fill=tk.BOTH, expand=True)
        vertical_split.add(path_inspector_frame)


        # Populate the canvas
        self._elements = {}
        for node in graph.nodes:
            self._elements[node] = GraphVertex(self, node)
            graph_frame.canvas.add_vertex(self._elements[node])

        for origin, edge, end in graph.edges:
            self._elements[(origin, edge, end)] = GraphEdge(self,
                                                      self._elements[origin],
                                                      self._elements[end],
                                                      edge)
            graph_frame.canvas.add_edge(self._elements[(origin, edge, end)])


        self.nodes_keys = get_all_keys(element.label_values
                                       for element in self._elements.values()
                                       if isinstance(element, GraphVertex))
        self.edges_keys = get_all_keys(element.label_values
                                       for element in self._elements.values()
                                       if isinstance(element, GraphEdge))

        # Mouses
        element_inspector_mouse = ElementInspectorMouse(self.
                                                        _inspected_element)
        graph_frame.canvas.register_mouse(element_inspector_mouse,
                                          button="1",
                                          modifiers="Alt")

        imm = InspectorModifyingMouse(self._inspected_path,
                                      graph_frame.canvas.edges)
        graph_frame.canvas.register_mouse(imm, button="1",
                                          modifiers="Control-Shift")

        pm = PathInspectingMouse(self._inspected_path,
                                 graph_frame.canvas.vertices,
                                 graph_frame.canvas.edges)
        graph_frame.canvas.register_mouse(pm, button="1", modifiers="Control")

        kcm = KeysChooserMouse(self)
        graph_frame.canvas.register_mouse(kcm, button="2")

        cmm = CustomMenuMouse()
        graph_frame.canvas.register_mouse(cmm, button="2")


        # Transformers
        def label_transformer(element, style):
            if isinstance(element, GraphVertex):
                keys = self.nodes_keys
            else: # GraphEdge
                keys = self.edges_keys

            keys = element_keys(element.label_values, follow=keys)
            keys.harmonize()

            element_label = label(element.label_values, keys)
            style["label"] = element_label
        graph_frame.canvas.register_transformer(label_transformer)

        def element_inspector_transformer(element, style):
            if element == self._inspected_element.value:
                style["shape_style"]["dash"] = 3
            else:
                style["shape_style"]["dash"] = ()
        graph_frame.canvas.register_transformer(element_inspector_transformer)

        def path_inspector_transformer(element, style):
            if (not self._inspected_path) or (element in self._inspected_path):
                style["shape_style"]["outline"] = "black"
                if isinstance(element, GraphEdge):
                    style["arrow_style"]["fill"] = "black"
            else:
                style["shape_style"]["outline"] = "gray"
                if isinstance(element, GraphEdge):
                    style["arrow_style"]["fill"] = "gray"
        graph_frame.canvas.register_transformer(path_inspector_transformer)
    
    @property
    def canvas(self):
        """
        The canvas displaying the graph.
        """
        return self._graph_frame.canvas
    
    @property
    def graph(self):
        """
        The displayed graph.
        """
        return self._graph
    
    @graph.setter
    def graph(self, new_graph):
        # Update the graph
        # Must update:
        #   - the graph itself
        #   - the canvas
        #   - the inspected element
        #   - the inspected path
        #   - the elements dictionary
        #   - the nodes keys
        #   - the edges keys
        
        def remove_element(element):
            """
            Remove the given element from all data structures
            (except the graph).
            
            element -- a displayed graph element.
            """
            self._graph_frame.canvas.delete_element(self._elements[element])
            if self._inspected_element.value == self._elements[element]:
                self._inspected_element.value = None
            self._inspected_path.discard(self._elements[element])
            del self._elements[element]
        
        def add_node(node):
            """
            Add the given node to all necessary data structures (except the
            graph).
            
            node -- a new node.
            """
            # Add the node in canvas and elements dictionary
            self._elements[node] = GraphVertex(self, node)
            self._graph_frame.canvas.add_vertex(self._elements[node])
        
        def add_edge(edge):
            """
            Add the given edge to all necessary data structures (except the
            graph).
            
            edge -- a new edge.
            """
            # Add the edge in canvas and elements dictionary
            origin, edge, end = edge
            self._elements[(origin, edge, end)] = GraphEdge(
                                                      self,
                                                      self._elements[origin],
                                                      self._elements[end],
                                                      edge)
            self._graph_frame.canvas.add_edge(self._elements[(origin,
                                                              edge,
                                                              end)])
        
        
        # Remove deleted edges
        for edge in set(self._graph.edges) - set(new_graph.edges):
            remove_element(edge)
        
        # Remove deleted nodes
        for node in set(self._graph.nodes) - set(new_graph.nodes):
            remove_element(node)
        
        # Add new nodes
        for node in set(new_graph.nodes) - set(self._graph.nodes):
            add_node(node)
        
        # Add new edges
        for edge in set(new_graph.edges) - set(self._graph.edges):
            add_edge(edge)
        
        # Update nodes keys and edges keys
        self.nodes_keys = get_all_keys((element.label_values
                                        for element in self._elements.values()
                                        if isinstance(element, GraphVertex)),
                                       follow=self.nodes_keys)
        self.nodes_keys.harmonize()
        self.edges_keys = get_all_keys((element.label_values
                                        for element in self._elements.values()
                                        if isinstance(element, GraphEdge)),
                                       follow=self.edges_keys)
        self.edges_keys.harmonize()
        
        # Update the graph
        self._graph = new_graph
    
    @property
    def inspected_element(self):
        """
        The currently inspected element (None if no element is inspected).
        The inspected element is the actual element of the inspected graph.
        """
        return self._inspected_element.value.data
    
    @inspected_element.setter
    def inspected_element(self, new_value):
        # new_value is an element of the graph
        self._inspected_element.value = self._elements[new_value]
    
    @property
    def inspected_path(self):
        """
        The currently inspected path (None if no path is inspected or the set
        of inspected elements does not define a path).
        This path is represented by a tuple of vertices and edges of the
        inspected graph, starting and ending with a vertex and alternating
        between vertices and edges.
        """
        path = _get_path(self._inspected_path)
        if path is None:
            return path
        else:
            return tuple(element.data for element in path)
    
    @inspected_path.setter
    def inspected_path(self, new_path):
        # new_path is an iterable of elements of the inspected graph
        self._inspected_path.clear()
        self._inspected_path |= {self._elements[element]
                                 for element in new_path}


def inspect(graph, parent=None, data=None):
    """
    Open an inspector for the given graph. If parent is not None,
    the new inspector window gets the given parent as parent.
    
    graph -- the graph to inspect;
    parent -- if not None, a tkinter widget.
    data -- any additional data to store in the inspector.
    """
    # Create window
    if parent is None:
        root = tk.Tk()
        mainloop = True
    else:
        root = tk.Toplevel(parent)
        mainloop = False
    
    inspector = InspectorFrame(graph, parent=root, data=data)
    
    if mainloop:
        root.mainloop()


"""
Relational Graph Algebra
Inspired from "Model Checking and Evidence Exploration",
by Y. Dong, C.R. Ramakrishnan, S.A. Smolka
(refered as [MCEE] in the sequel)

A domain is either a primitive domain (integers, reals, etc.)
or a Cartesian product of domains D_1, ..., D_n. [MCEE]

A relational graph is a directed graph <V, E> where vertices V are elements
of the node domain D_V and edges E are elements of V x D_E x V, where D_E
is the edge domain. [MCEE]

The basic operators are [MCEE]:
 * the union of two graphs defined on the same node and edge domains;
 * the intersection of two graphs defined on the same node and edge domains;
 * the node difference of two graphs defined on the same node and edge
   domains;
 * the edge difference of two graphs defined on the same node and edge
   domains;
 * the selection of a part of a graph based on two boolean functions for
   selecting nodes and edges, respectively;
 * the projection of a graph on two node and edge sub-domains;
 * the Cartesian product of two graphs.


In terms of implementation, a Graph has a set of nodes and a set of edges, that
is, a set of triplets where the first and third elements are nodes and the
second element is the edge value.

Nodes and edge values must be domaintuples. The domain of a node or edge is
the set of keys of the corresponding domaintuple. Domaintuples values MUST
be hashable.

Nodes and edges do not have to share the same domains, as in [MCEE]; instead,
the full domain of nodes and edges is the union of all the keys of the
domaintuples of nodes and edges, respectively.
"""

from collections import Mapping, defaultdict

class domaintuple(Mapping):
    """
    A domaintuple is a read-only dictionary where missing keys default to None.
    getitem will thus never fails and will return None if the key is not
    present; keys and items are only the ones where the value is note None.
    At construction, all None values are removed from the initializing values
    such that only non-None values are kept.

    Domaintuples values MUST be hashable.

    Inspired from frozendict from https://pypi.python.org/pypi/frozendict/.
    """

    def __init__(self, *args, **kwargs):
        self.__dict = {k: v
                       for k, v in dict(*args, **kwargs).items()
                       if v is not None}
        self.__hash = None

    def __getitem__(self, key):
        try:
            return self.__dict[key]
        except KeyError:
            return None
    
    def __getattr__(self, name):
        if name in self.__dict:
            return self.__dict[name]
        else:
            return None

    def copy(self):
        return domaintuple(self)

    def __iter__(self):
        return iter(self.__dict)

    def __len__(self):
        return len(self.__dict)

    def __repr__(self):
        return '<domaintuple %s>' % repr(self.__dict)

    def __hash__(self):
        if self.__hash is None:
            self.__hash = hash(frozenset(self.__dict.items()))
        return self.__hash

    def keys(self):
        return self.__dict.keys()

    def values(self):
        return self.__dict.values()

    def items(self):
        return self.__dict.items()

    def __contains__(self, key):
        return key in self.__dict

    def get(self, key, default=None):
        return self.__dict.get(key, default=default)

    def __eq__(self, other):
        return self.__dict == other

    def __ne__(self, other):
        return self.__dict != other


class Graph(Mapping):
    """
    A relational graph.
    
    A graph is represented by a mapping, giving, for each node, the set of
    pairs of edge value and successing node. This set of pairs must be a
    frozenset and each pair must be a tuple.
    
    A constraint any graph must fulfill is that each node appearing as a
    successor must also appear as a key of the mapping (even if its set of
    successing nodes is empty).
    """
    
    def __init__(self, *args, **kwargs):
        """
        Create a new relational graph. Its arguments must initialize the
        mapping corresponding to the graph.
        """
        self.__dict = dict(*args, **kwargs)
        assert(self.check())
    
    @property
    def nodes(self):
        return self.__dict.keys()
    
    @property
    def edges(self):
        for node, successors in self.items():
            for edge, successor in successors:
                yield (node, edge, successor)
    
    def check(self):
        """
        Check whether this graph respects the constraints of graphs:
         * each key is a domaintuple
         * each value is a frozenset of pairs of domaintuples
         * each successor is a key of this graph
        """
        for node, successors in self.items():
            if not isinstance(node, domaintuple):
                return False
            if not isinstance(successors, frozenset):
                return False
            for edge, successor in successors:
                if not isinstance(edge, domaintuple):
                    return False
                if not isinstance(successor, domaintuple):
                    return False
                if successor not in self:
                    return False
        return True
    
    def __getitem__(self, key):
        return self.__dict[key]

    def copy(self):
        return Graph(self)

    def __iter__(self):
        return iter(self.__dict)

    def __len__(self):
        return len(self.__dict)

    def __repr__(self):
        return '<Graph %s>' % repr(self.__dict)

    def union(self, other):
        """
        Compute the union of this graph and other.

        other -- another Graph.

        Return a new Graph representing the union of self and other.
        """
        new_graph = defaultdict(frozenset, self)
        for node, successors in other.items():
            new_graph[node] |= successors
        return Graph(new_graph)

    __or__ = union

    def intersection(self, other):
        """
        Compute the intersection of this graph and other.

        other -- another Graph.

        Return a new Graph representing the intersection of self and other.
        """
        new_graph = {node: frozenset({(edge, successor)
                                      for edge, successor in successors
                                      if (edge, successor) in other[node]})
                     for node, successors in self.items()
                     if node in other}
        return Graph(new_graph)

    __and__ = intersection

    def node_difference(self, other):
        """
        Compute the node difference of this graph and other.

        other -- another Graph.

        Return a new Graph representing the node difference of self and other.
        """
        new_graph = {node: frozenset({(edge, successor)
                                      for edge, successor in successors
                                      if successor not in other})
                     for node, successors in self.items()
                     if node not in other}
        return Graph(new_graph)

    def edge_difference(self, other):
        """
        Compute the edge difference of this graph and other.

        other -- another Graph.

        Return a new Graph representing the edge difference of self and other.
        """
        new_graph = {node: frozenset({(edge, successor)
                                      for edge, successor in successors
                                      if node not in other or
                                         (edge, successor) not in other[node]})
                     for node, successors in self.items()}
        return Graph(new_graph)

    def selection(self, node_selector=None, edge_selector=None):
        """
        Compute the selection of this graph according to node_selector and
        edge_selector.

        node_selector -- a function taking a node as argument and returning
                         whether or not it is selected or not;
        edge_selector -- a function taking an edge as argument and
                         returning whether or not it is selected or not.

        If node_selector is None, all nodes are kept; if edge_selector is None,
        all edges for which the extremities are kept, are kept.

        Return a new Graph where nodes are all nodes of this graph for which
        node_selector returns a true value, and where edges are all edges of
        this graph for which edge_selector returns a true value.
        """
        if node_selector is None:
            nodes = set(self.nodes)
        else:
            nodes = {node for node in self if node_selector(node)}
        new_graph = {node: frozenset({(edge, successor)
                                      for edge, successor in successors
                                      if successor in nodes and
                                         (edge_selector is None or
                                          edge_selector((node,
                                                         edge,
                                                         successor)))})
                     for node, successors in self.items()
                     if node in nodes}
        return Graph(new_graph)

    def projection(self, node_domain=None, edge_domain=None):
        """
        Compute the projection of this graph on the given node and edge
        domains.

        node_domain -- a sub-domain of the node domain of this graph, that is,
                       a set of domain names;
        edge_domain -- a sub-domain of the edge domain of this graph, that is,
                       a set of domain names.

        If node_domain is None, nodes are kept as they are; if edge_domain is
        None, edge values are kept as they are.

        Return a new Graph where nodes are new domaintuples with keys of
        node_domain and edges are new domaintuples with keys of edge_domain.
        Edges without non-None values are removed if edge_domain is given.
        """
        if node_domain is None:
            nodes = {node: node for node in self.nodes}
            new_graph = {node: set() for node in nodes}
        else:
            nodes = {node: self.project(node, node_domain)
                     for node in self.nodes}
            new_graph = {node: set() for node in nodes.values()}
        if edge_domain is None:
            for origin, edge, end in self.edges:
                new_graph[nodes[origin]].add((edge, nodes[end]))
        else:
            for origin, edge, end in self.edges:
                projected = self.project(edge, edge_domain)
                if projected:
                    new_graph[nodes[origin]].add((projected, nodes[end]))
        return Graph({node: frozenset(successors)
                      for node, successors in new_graph.items()})

    def join(self, other):
        """
        Compute the join of this graph with other.

        other -- another Graph.

        Return a new Graph where the node set is the join of this graph
        nodes and the other graph nodes, that is, each node of this graph
        is joined with each node of the other graph for which the common
        domains have the same value,
        and each edge embeds at least one edge from this graph or the other
        graph, and keeps the remaining idle (None values), that is,
        the resulting edge values are each edge value of this graph with None
        values for the domains of the other graph and vice versa, plus the
        joined edge values, that is, the edge values that agree on common
        domains.
        """
        new_graph = {}
        nodes = {}
        # A dictionary of pairs of nodes telling whether they agree on common
        # domains
        agree = {}
        for node1 in self.nodes:
            for node2 in other.nodes:
                if (node1, node2) not in agree:
                    agree[(node1, node2)] = all(node1[key] == node2[key]
                                                for key in node1.keys() &
                                                           node2.keys())
                if agree[(node1, node2)]:
                    nodes[(node1, node2)] = self.join_elements(node1, node2)
                    new_graph[nodes[(node1, node2)]] = set()

        for origin1, edge1, end1 in self.edges:
            for origin2, edge2, end2 in other.edges:
                if (agree[(origin1, origin2)] and
                    agree[(end1, end2)]):
                    if (edge1, edge2) not in agree:
                        agree[(edge1, edge2)] = all(edge1[key] == edge2[key]
                                                    for key in edge1.keys() &
                                                               edge2.keys())
                    if agree[(edge1, edge2)]:
                        new_graph[nodes[(origin1, origin2)]].add(
                            (self.join_elements(edge1, edge2),
                             nodes[(end1, end2)]))
        for origin1, edge1, end1 in self.edges:
            for node2 in other.nodes:
                if (agree[(origin1, node2)] and
                    agree[(end1, node2)]):
                    new_graph[nodes[(origin1, node2)]].add(
                        (edge1, nodes[(end1, node2)]))
        for node1 in self.nodes:
            for origin2, edge2, end2 in other.edges:
                if (agree[(node1, origin2)] and
                    agree[(node1, end2)]):
                    new_graph[nodes[(node1, origin2)]].add(
                        (edge2, nodes[(node1, end2)]))
        return Graph({node: frozenset(successors)
                      for node, successors in new_graph.items()})

    def extension(self, node_extension=None, edge_extension=None):
        """
        Extend each node with node_extension and each edge value with
        edge_extension.

        node_extension -- a function taking a node as argument and returning
                          new domain values for this node;
        edge_extension -- a function taking an edge (origin, edge value, end)
                          as argument and returning new domain values for this
                          edge.

        If node_extension is None, nodes are kept as they are;
        if edge_extension is None, edges are kept as they are.

        node_extension and edge_extension can return new values for sub-domains
        for which nodes and edges are already defined and the new values will
        replace the old ones.
        """
        new_graph = {}
        if node_extension is not None:
            nodes = {}
            for node in self.nodes:
                nodes[node] = self.extend(node, node_extension(node))
                new_graph[nodes[node]] = set()
        else:
            nodes = {node: node for node in self.nodes}
            new_graph = {node: set() for node in self.nodes}

        for origin, edge, end in self.edges:
            if edge_extension is not None:
                new_graph[nodes[origin]].add(
                    (self.extend(edge, edge_extension((origin, edge, end))),
                     nodes[end]))
            else:
                new_graph[nodes[origin]].add((domaintuple(edge), nodes[end]))

        return Graph({node: frozenset(successors)
                      for node, successors in new_graph.items()})

    def mapping(self, node_mapping=None, edge_mapping=None):
        """
        Map each node to node_mapping and each edge value to edge_mapping.

        node_mapping -- a function taking a node as argument and returning the
                        corresponding new node;
        edge_mapping -- a function taking an edge as argument (origin,
                        edge value, end) and returning the corresponding edge
                        value.

        If node_mapping is None, all nodes are kept as they are;
        if edge_mapping is None, all edge values are kept as they are.
        """
        if node_mapping is not None:
            nodes = {node: domaintuple(node_mapping(node))
                     for node in self.nodes}
            new_graph = {node: set() for node in nodes.values()}
        else:
            nodes = {node: node for node in self.nodes}
            new_graph = {node: set() for node in nodes}
        for origin, edge, end in self.edges:
            if edge_mapping is not None:
                new_edge = domaintuple(edge_mapping((origin, edge, end)))
            else:
                new_edge = edge
            new_graph[nodes[origin]].add((new_edge, nodes[end]))
        return Graph({node: frozenset(successors)
                      for node, successors in new_graph.items()})

    def grouping(self, node_group=None, edge_group=None):
        """
        Group all nodes with the same values for domains in the given node
        domains and accumulate domains outside the given domains into
        a new domain named with the given name;
        all edges between two grouped nodes are grouped according to the same
        rules.

        node_group -- a couple where the first element is the name for the new
                      domain in which the groups are given and the second
                      element is the set of domain names used for grouping;
        edge_group -- a couple where the first element is the name for the new
                      domain in which the groups are given and the second
                      element is the set of domain names used for grouping.

        If node_group is None, no grouping is made on nodes; if edge_group is
        None, no grouping is made on edge values.
        """
        if node_group is not None:
            node_group_name, node_domains = node_group
            grouped_nodes = defaultdict(lambda: set())
            for node in self.nodes:
                grouped_nodes[self.project(node, node_domains)].add(node)
            nodes = {}
            for projected in grouped_nodes:
                inner = frozenset(self.project(n, n.keys() - node_domains)
                                  for n in grouped_nodes[projected])
                grouped_node = self.extend(projected, {node_group_name: inner})
                for node in grouped_nodes[projected]:
                    nodes[node] = grouped_node
        else:
            nodes = {node: node for node in self.nodes}

        new_graph = {node: set() for node in nodes.values()}

        if edge_group is not None:
            edge_group_name, edge_domains = edge_group
            grouped_edges = defaultdict(lambda: set())
            for origin, edge, end in self.edges:
                grouped_edges[(nodes[origin],
                               self.project(edge, edge_domains),
                               nodes[end])].add(edge)
            edges = set()
            for origin, projected, end in grouped_edges:
                inner = frozenset(self.project(e, e.keys() - edge_domains)
                                  for e in grouped_edges[(origin,
                                                          projected,
                                                          end)])
                grouped_edge = self.extend(projected, {edge_group_name: inner})
                edges.add((origin, grouped_edge, end))
        else:
            edges = {(nodes[origin], edge, nodes[end])
                     for origin, edge, end in self.edges}
        for origin, edge, end in edges:
            new_graph[origin].add((edge, end))
        return Graph({node: frozenset(successors)
                      for node, successors in new_graph.items()})

    def ungrouping(self, node_group=None, edge_group=None):
        """
        Ungroup all nodes and edges based on the given domains.
        Duplicate each node and edge and replace the value of node_group (and
        edge_group, respectively) by each value it contains; if the value of
        node_group (or edge_group) is not iterable, it is kept as it is.

        node_group -- if not None, the name of a domain of nodes;
        edge_group -- if not None, the name of a domain of edges.

        If node_group is None, no ungrouping is made on nodes; if edge_group is
        None, no ungrouping is made on edges.
        """
        if node_group is None:
            nodes = {node: {node} for node in self.nodes}
        else:
            nodes = {node: self.ungroup(node, node_group)
                     for node in self.nodes}
        new_graph = {}
        for node, successors in self.items():
            for new_node in nodes[node]:
                new_graph[new_node] = set()
                for edge, successor in successors:
                    for new_edge in ({edge} if edge_group is None
                                     else self.ungroup(edge, edge_group)):
                        for new_succ in nodes[successor]:
                            new_graph[new_node].add((new_edge, new_succ))
        return Graph({node: frozenset(successors)
                      for node, successors in new_graph.items()})

    def dot(self):
        """
        Return a dot representation of this graph.

        The nodes and edges of the returned dot graph are labelled with their
        domain values (as "key = value"), sorted alphabetically.

        If a node or edge value has a domain named "dot", its value is used
        as the style of the corresponding node or edge.
        """
        ids = {}
        dot = "digraph {\n"

        nodes = []
        for nid, node in enumerate(self.nodes):
            ids[node] = nid
            if "dot" in node:
                style = node["dot"]
            else:
                label = "\n".join(str(key) + " = " + str(value)
                                  for key, value in sorted(node.items()))
                style = 'label="%s"' % label
            nodes.append('s%d [%s]' % (ids[node], style))
        dot += "\n".join(nodes) + "\n"

        edges = []
        for origin, edge, end in self.edges:
            if "dot" in edge:
                style = edge["dot"]
            else:
                label = "\n".join(str(key) + " = " + str(value)
                                  for key, value in sorted(edge.items()))
                style = 'label="%s"' % label
            edges.append('s%d -> s%d [%s]' %
                         (ids[origin], ids[end], style))
        dot += "\n".join(edges)

        dot += "\n}"
        return dot

    @staticmethod
    def ungroup(element, domain):
        """
        Ungroup domain value of element into new domaintuples for each value of
        domain.
        That is, for each value of domain in element, a new domaintuple is
        created, sharing the same other values as element, and getting only one
        of the values of domain in element as domain.
        If the value of domain in element is an iterable of domaintuples, the
        domains of these tuples are flattened in element instead of getting one
        new value containing the whole domaintuple.
        
        element -- a domaintuple;
        domain -- a domain of element.
        
        If the value of domain in element is not iterable, it is kept as it is
        and element is returned.
        The returned value is the set of new elements, even if domain in
        element is not iterable (in this case, the returned set contains only
        one element).
        """
        new_elements = set()
        try:
            for value in element[domain]:
                if isinstance(value, domaintuple):
                    old = dict(element)
                    del old[domain]
                    old.update(value)
                    new_elements.add(domaintuple(old))
                else:
                    new_elements.add(Graph.extend(element, {domain: value}))
        except TypeError:
            new_elements = {element}
        return new_elements

    @staticmethod
    def group(element, elements, domain, group_name):
        """
        Return the grouped domaintuple of all elements sharing the same domain
        values as element.

        element -- a domaintuple;
        elements -- a set of domaintuple;
        domain -- a sub-domain of element's domain.
        
        Return a domaintuple where domains in the given sub-domain are the same
        as element, and where the sub-domain outside the given sub-domain
        is the accumulation of all the other domaintuples in elements sharing
        the same value as element in the given sub-domain.
        """
        outer = dict(Graph.project(element, domain))
        inner = set()
        for other in elements:
            if Graph.project(other, domain) == outer:
                inner.add(Graph.project(other, other.keys() - domain))
        outer.update({group_name: frozenset(inner)})
        return domaintuple(outer)

    @staticmethod
    def project(element, domain):
        """
        Return the projection of the given element on the given domain.

        element -- a domaintuple;
        domain -- a subset of element keys.

        Return a new domaintuple where keys are the elements of domain and
        values are the corresponding values in element.
        """
        return domaintuple({k: v for k, v in element.items()
                                 if k in domain})

    @staticmethod
    def join_elements(element1, element2):
        """
        Return the join of the two elements.

        element1 -- a domaintuple;
        element2 -- another domaintuple; agrees on domains common with
                    element1.

        Return a new domaintuple with all the keys of element1 and element2,
        and the corresponding values in the corresponding elements
        (note that common keys agree).
        """
        element = dict(element1)
        element.update(element2)
        return domaintuple(element)

    @staticmethod
    def extend(old, new):
        """
        Return the extension of old with new.

        old -- a domaintuple;
        new -- another domaintuple; defined on a disjoint domain.

        Return a new domaintuple defined on the union of domains of old and new
        and agreeing with the corresponding domaintuples on their respective
        domains.
        """
        old = dict(old)
        old.update(new)
        return domaintuple(old)


"""
Functions to compute mu-calculus models with NuSMV.
"""

from pynusmv_lower_interface.nusmv.compile.symb_table import symb_table as nssymb_table
from pynusmv_lower_interface.nusmv.compile import compile as nscompile
from pynusmv_lower_interface.nusmv.node import node as nsnode

from pynusmv.exception import NuSMVNoReadModelError
from pynusmv.parser import parse_simple_expression, parse_next_expression

from pynusmv.glob import (load_from_file, load_from_string,
                          load_from_modules, load,
                          flatten_hierarchy as _flatten_hierarchy,
                          symb_table, bdd_encoding,
                          prop_database as _prop_database,
                          compute_model as _compute_model)
from pynusmv.mc import eval_simple_expression
from pynusmv.fsm import BddFsm, BddTrans

from . import MuCalculusError, MuCalculusModelError, UnknownTransitionError

import itertools


class InputsBasedBddTrans(BddTrans):
    """
    A BDD-encoded transition relation based on a classical BDD-encoded
    transition relation and a given inputs. Every transition evaluation is
    restricted by this inputs.
    
    The idea behind this inputs is to be able to define different transition
    relations based on the same definition but with additional restrictions.
    """
    
    def __init__(self, ptr, inputs, enc=None, manager=None, freeit=True):
        """
        Create a new InputsBasedBddTrans.

        :param ptr: a NuSMV pointer to a BddTrans
        :param inputs: a BDD representing the inputs retricting this transition
                       relation
        :param enc: the BDD encoding of the transition relation
        :type enc: :class:`BddEnc`
        :param manager: the DD manager of the BDDs used to encode the relation
        :type manager: :class:`DDManager <pynusmv.dd.DDManager>`
        :param freeit: whether or not freeing the pointer

        """
        super(InputsBasedBddTrans, self).__init__(ptr, enc=enc,
                                                  manager=manager,
                                                  freeit=freeit)
        self._inputs = inputs
    
    def pre(self, states, inputs=None):
        """
        Compute the pre-image of `states`, through `inputs` if not `None`.

        :param states: the concerned states
        :type states: :class:`BDD <pynusmv.dd.BDD>`
        :param inputs: possible inputs
        :type inputs: :class:`BDD <pynusmv.dd.BDD>`
        :rtype: :class:`BDD <pynusmv.dd.BDD>`

        """
        inputs = self._inputs & inputs if inputs else self._inputs
        return super(InputsBasedBddTrans, self).pre(states, inputs=inputs)

    def post(self, states, inputs=None):
        """
        Compute the post-image of `states`, through `inputs` if not `None`.

        :param states: the concerned states
        :type states: :class:`BDD <pynusmv.dd.BDD>`
        :param inputs: possible inputs
        :type inputs: :class:`BDD <pynusmv.dd.BDD>`
        :rtype: :class:`BDD <pynusmv.dd.BDD>`

        """
        inputs = self._inputs & inputs if inputs else self._inputs
        return super(InputsBasedBddTrans, self).post(states, inputs=inputs)
    
    @classmethod
    def from_trans(cls, symb_table, inputs, trans, context=None):
        """
        Return a new ReversedBddTrans from the given trans.

        :param symb_table: the symbols table used to flatten the trans
        :type symb_table: :class:`SymbTable`
        :param trans: the parsed string of the trans, not flattened
        :param inputs: the inputs to restrict this transition relation
        :param context: an additional parsed context, in which trans will be
                        flattened, if not None
        :rtype: :class:`BddTrans`
        :raise: a :exc:`NuSMVFlatteningError
                <pynusmv.exception.NuSMVFlatteningError>`
                if `trans` cannot be flattened under `context`

        """
        original = BddTrans.from_trans(symb_table, trans, context)
        return InputsBasedBddTrans(original._ptr, inputs,
                                   original._enc, original._manager,
                                   freeit=True)


class BddMuModel(BddFsm):
    """
    The BDD representation of a mu-calculus model.
    
    A mu-calculus model is an FSM with sereval transition relations.
    """
    
    def __init__(self, ptr, transitions, freeit=False):
        """
        Create a new mu-calculus model.
        
        ptr -- the pointer of the NuSMV FSM
        transitions -- a dictionary of transition names -> BddTrans.
        freeit -- whether or not free the pointer
        """
        super(BddMuModel, self).__init__(ptr, freeit = freeit)
        self.transitions = transitions
    
    
    def pre(self, states, transition="time", inputs=None):
        """
        Return the pre-image of states in this model through transition.
        
        If inputs is not None, it is used to restrict the pre-image
        computation.
        
        states -- the BDD representing the set of states;
        transition -- the transition name;
        inputs -- if not None, a BDD on inputs variables, restricting the
                  pre-image.
        
        """
        if transition not in self.transitions:
            raise UnknownTransitionError("Unknown transition " +
                                         str(transition))
        return self.transitions[transition].pre(states, inputs=inputs)
    
    
    def post(self, states, transition="time", inputs=None):
        """
        Return the post-image of states in this model through transition.
        
        If inputs is not None, it is used to restrict the post-image
        computation.
        
        states -- the BDD representing the set of states;
        transition -- the transition name;
        inputs -- if not None, a BDD on inputs variables, restricting the
                  post-image.
        
        """
        if transition not in self.transitions:
            raise UnknownTransitionError("Unknown transition " + transition)
        return self.transitions[transition].post(states, inputs=inputs)


# The current model
__bddmodel = None

def reset_globals():
    """
    Reset mucalculus.model related global variables.
    
    Must be called whenever (and before) pynusmv.init.init.deinit_nusmv
    is called.
    
    """
    global __bddmodel
    __bddmodel = None


def bddModel(transitions=None, variables_ordering=None):
    """
    Return (and compute if needed) the model represented by the currently read
    SMV model.
    
    transitions -- if not None, a dictionary of transition relation names ->
                   transition relations expressions as strings.
    variables_ordering -- if not None, a path to a variables order file.
    
    The currently read SMV model must be defined in terms of an input variable
    named 'transition' with an enum type. All the possible values of this
    'transition' input variable are considered as different transition
    relations and the returned BddModel is based on them.
    
    If such a transition input variable is not defined, the standard transition
    relation of the model is used and named "time".
    
    If transitions is not None, additional transition relations a built and
    added to the return BddModel.
    
    """    
    global __bddmodel
    if __bddmodel is None:
        # Check cmps
        if not nscompile.cmp_struct_get_read_model(nscompile.cvar.cmps):
            raise NuSMVNoReadModelError("Cannot build model; no read file.")        
        # Compute the model
        _compute_model(variables_ordering=variables_ordering)
        
        # Check that transition exists and is an input variable enum
        # Get transition node
        transition = parse_simple_expression("transition")
        st = symb_table()
        
        # Resolve name and get category
        rs = nssymb_table.SymbTable_resolve_symbol(st._ptr, transition, None)
        tr_name = nssymb_table.ResolveSymbol_get_resolved_name(rs)
        if not nssymb_table.ResolveSymbol_is_defined(rs):
            # Transition is missing
            # Create the model
            fsm = _prop_database().master.bddFsm
            # Create embedded transition relations
            relations = {"time": fsm.trans}
        else:
            tr_category = nssymb_table.SymbTable_get_symbol_category(st._ptr,
                                                                     tr_name)
            
            # If constant, check that is a define and get value
            if tr_category == nssymb_table.SYMBOL_CONSTANT:
                # Check transition is a define
                if not nssymb_table.SymbTable_is_symbol_define(st._ptr,
                                                               tr_name):
                    raise MuCalculusModelError("Transition should be an input "
                                               "variable.")
                tr_value = (nssymb_table.
                            SymbTable_get_define_flatten_body(st._ptr,
                                                              tr_name))
                tr_names = {nsnode.sprint_node(tr_value)}
                
            # If input var, check that is an enum
            elif tr_category == nssymb_table.SYMBOL_INPUT_VAR:
                tr_type = nssymb_table.SymbTable_get_var_type(st._ptr, tr_name)
                if not nssymb_table.SymbType_is_enum(tr_type):
                    raise MuCalculusModelError("Transition should be an enum "
                                               "type.")
                tr_values = nssymb_table.SymbType_get_enum_type_values(tr_type)
                tr_names = set()
                while tr_values is not None:
                    tr_value = nsnode.car(tr_values)
                    tr_names.add(nsnode.sprint_node(tr_value))
                    tr_values = nsnode.cdr(tr_values)
            else:
                raise MuCalculusModelError("Transition should be an input "
                                           "variable.")
            
            # Create the model
            fsm = _prop_database().master.bddFsm
            
            # Create embedded transition relations
            relations = {}
            original_trans = fsm.trans
            enc = original_trans._enc
            manager = original_trans._manager
            for tr_name in tr_names:
                inputs = eval_simple_expression(fsm, 'transition =' + tr_name)
                relations[tr_name] = InputsBasedBddTrans(original_trans._ptr,
                                                         inputs, enc=enc,
                                                         manager=manager,
                                                         freeit=False)
        
        # Create additional transition relations
        if transitions is not None:
            for tr_name, tr_expr in transitions.items():
                relations[tr_name] = BddTrans.from_string(st, tr_expr)
        
        __bddmodel = BddMuModel(fsm._ptr, relations, freeit=False)
        
    return __bddmodel


from pyparsing import Suppress, SkipTo, Forward, ZeroOrMore, Literal, Group

from .ast import (TrueExp, FalseExp, Init, Reachable,
                  Atom, Not, And, Or, Implies, Iff, 
                  AF, AG, AX, AU, AW, EF, EG, EX, EU, EW,
                  nK, nE, nD, nC, K, E, D, C)
                  
                  
def _left_(clss, tokens):
    """
    Parse tokens and return an AST, assuming left associativity.
    
    Given a list of tokens [v1, op, v2, op, ..., op, vn],
    return res, a hierarchy of instances of clss such that
    res = clss(clss(...clss(v1, v2), ..., vn).
    
    This is a helper function to parse logical operators.
    """
    if len(tokens) == 1:
        return tokens[0]
    else:
        return clss(_left_(clss, tokens[:-2]), tokens[-1])


def _right_(clss, tokens):
    """
    Parse tokens and return an AST, assuming right associativity.
    
    Given a list of tokens [v1, op, v2, op, ..., op, vn],
    return res, a hierarchy of instances of clss such that
    res = clss(v1, clss(v2, ..., vn)...).
    
    This is a helper function to parse logical operators.
    """
    if len(tokens) == 1:
        return tokens[0]
    else:
        return clss(tokens[0], _right_(clss, tokens[2:]))
        
        
def _logicals_(atomic):
    """
    Return a new parser parsing logical expressions on atomics.
    
    This parser recognizes the following grammar, with precedences
    parser := atomic | '~' parser | parser '&' parser | parser '|' parser |
              parser '->' parser | parser '<->' parser
              
    Returned AST uses .ast package's classes.
    """
    parser = Forward()

    atom = (atomic | Suppress("(") + parser + Suppress(")"))

    notstrict = "~" + atom
    notstrict.setParseAction(lambda tokens: Not(tokens[1]))
    not_ = notstrict | atom
    and_ = not_ + ZeroOrMore("&" + not_)
    and_.setParseAction(lambda tokens: _left_(And, tokens))
    or_ = and_ + ZeroOrMore("|" + and_)
    or_.setParseAction(lambda tokens: _left_(Or, tokens))
    implies = ZeroOrMore(or_ + "->") + or_
    implies.setParseAction(lambda tokens: _right_(Implies, tokens))
    iff = implies + ZeroOrMore("<->" + implies)
    iff.setParseAction(lambda tokens: _left_(Iff, tokens))

    parser <<= iff
    
    return parser
        

"""
ARCTL parsing tool.

__ctlk       := _atom | _logical | _temporal | _epistemic
_logical    := '~' __ctlk | '(' _logical ')' | __ctlk '&' __ctlk |
               __ctlk '|' __ctlk | __ctlk '->' __ctlk | __ctlk '<->' __ctlk
_temporal   := 'A' 'F' __ctlk | 'A' 'G' __ctlk | 'A' 'X' __ctlk |
               'A' '[' __ctlk 'U' __ctlk ']' | 'A' '[' __ctlk 'W' __ctlk ']' |
               'E' 'F' __ctlk | 'E' 'G' __ctlk | 'E' 'X' __ctlk |
               'E' '[' __ctlk 'U' __ctlk ']' | 'E' '[' __ctlk 'W' __ctlk ']'
_epistemic  := 'nK' '<' _agent '>' __ctlk | 'nE' '<' _group '>' __ctlk |
               'nD' '<' _group '>' __ctlk | 'nC' '<' _group '>' __ctlk |
               'K' '<' _agent '>' __ctlk | 'E' '<' _group '>' __ctlk |
               'D' '<' _group '>' __ctlk | 'C' '<' _group '>' __ctlk
_agent      := _atom
_group      := _agent | _agent ',' _group

               
_atom is defined by any string surrounded by single quotes.

_logical are specified with usual precedences and associativity,
i.e.
prec : ~, &, |, ->, <->
assoc : &, |, <-> left assoc, ->, ~ right assoc


The parser returns a structure embedding the structure of the parsed
expression, represented using AST classes of .ast module.
"""

__ctlk = None

def parseCTLK(spec):
    """Parse the spec and return the list of possible ASTs."""
    global __ctlk
    if __ctlk is None:
        true = Literal("True")
        true.setParseAction(lambda tokens: TrueExp())
        false = Literal("False")
        false.setParseAction(lambda tokens: FalseExp())
        init = Literal("Init")
        init.setParseAction(lambda tokens: Init())
        reachable = Literal("Reachable")
        reachable.setParseAction(lambda tokens: Reachable())
        
        atom = "'" + SkipTo("'") + "'"
        atom.setParseAction(lambda tokens: Atom(tokens[1]))
        
        agent = atom
        group = Group(ZeroOrMore(agent + Suppress(",")) + agent)
        
        proposition = true | false | init | reachable | atom
        
        __ctlk = Forward()

        notproposition = "~" + proposition
        notproposition.setParseAction(lambda tokens: Not(tokens[1]))
        formula = (proposition | notproposition |
                   Suppress("(") + __ctlk + Suppress(")"))

        logical = Forward()
        

        ex = Literal("E") + "X" + logical
        ex.setParseAction(lambda tokens: EX(tokens[2]))
        ax = Literal("A") + "X" + logical
        ax.setParseAction(lambda tokens: AX(tokens[2]))
         
        ef = Literal("E") + "F" + logical
        ef.setParseAction(lambda tokens: EF(tokens[2]))
        af = Literal("A") + "F" + logical
        af.setParseAction(lambda tokens: AF(tokens[2]))
         
        eg = Literal("E") + "G" + logical
        eg.setParseAction(lambda tokens: EG(tokens[2]))
        ag = Literal("A") + "G" + logical           
        ag.setParseAction(lambda tokens: AG(tokens[2]))
         
        eu = Literal("E") + "[" + __ctlk + "U" + __ctlk + "]"
        eu.setParseAction(lambda tokens: EU(tokens[2], tokens[4]))
        au = Literal("A") + "[" + __ctlk + "U" + __ctlk + "]"   
        au.setParseAction(lambda tokens: AU(tokens[2], tokens[4]))
                                                                
        ew = Literal("E") + "[" + __ctlk + "W" + __ctlk + "]"   
        ew.setParseAction(lambda tokens: EW(tokens[2], tokens[4]))
        aw = Literal("A") + "[" + __ctlk + "W" + __ctlk + "]"   
        aw.setParseAction(lambda tokens: AW(tokens[2], tokens[4]))

        temporal = (ex | ax | ef | af | eg | ag | eu | au | ew | aw)
        
        
        nk = Literal("nK") + "<" + agent + ">" + logical
        nk.setParseAction(lambda tokens: nK(tokens[2], tokens[4]))
        k = Literal("K") + "<" + agent + ">" + logical
        k.setParseAction(lambda tokens: K(tokens[2], tokens[4]))
        
        ne = Literal("nE") + "<" + group + ">" + logical
        ne.setParseAction(lambda tokens: nE(list(tokens[2]), tokens[4]))
        e = Literal("E") + "<" + group + ">" + logical
        e.setParseAction(lambda tokens: E(list(tokens[2]), tokens[4]))
        
        nd = Literal("nD") + "<" + group + ">" + logical
        nd.setParseAction(lambda tokens: nD(list(tokens[2]), tokens[4]))
        d = Literal("D") + "<" + group + ">" + logical
        d.setParseAction(lambda tokens: D(list(tokens[2]), tokens[4]))
        
        nc = Literal("nC") + "<" + group + ">" + logical
        nc.setParseAction(lambda tokens: nC(list(tokens[2]), tokens[4]))
        c = Literal("C") + "<" + group + ">" + logical
        c.setParseAction(lambda tokens: C(list(tokens[2]), tokens[4]))
        
        epistemic = (nk | k | ne | e | nd | d | nc | c)
        
        logical <<= (formula | epistemic | temporal)

        __ctlk <<= (_logicals_(logical))
    
    return __ctlk.parseString(spec, parseAll = True)